# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""
Configuration
-------------

Handle *plotpy* module configuration
(options, images and icons)
"""

from __future__ import annotations

import os.path as osp
import warnings
from typing import Literal

from guidata import qthelpers as qth
from guidata.configtools import add_image_module_path, get_translation
from guidata.userconfig import UserConfig


def get_main_colors() -> tuple[str, str]:
    """Return main foreground and background colors depending on dark mode"""
    is_dark = qth.is_dark_theme()
    return "white" if is_dark else "black", "#222222" if is_dark else "white"


MAIN_FG_COLOR, MAIN_BG_COLOR = get_main_colors()


def make_title(basename, count):
    """Make item title with *basename* and *count* number"""
    return "{} {}{:d}".format(basename, _("#"), count)


APP_PATH = osp.dirname(__file__)
add_image_module_path("plotpy", "data/icons")
_ = get_translation("plotpy")


def get_plotpy_defaults() -> dict[str, int | float | str | bool]:
    """Return default configuration values"""
    global MAIN_FG_COLOR, MAIN_BG_COLOR
    return {
        "plot": {
            "selection/distance": 6,
            "antialiasing": False,
            "title/font/size": 12,
            "title/font/bold": False,
            "selected_curve_symbol/marker": "Rect",
            "selected_curve_symbol/edgecolor": "gray",
            "selected_curve_symbol/facecolor": MAIN_FG_COLOR,
            "selected_curve_symbol/alpha": 0.3,
            "selected_curve_symbol/size": 7,
            # Default parameters for plot axes
            "axis/title": "",
            "axis/unit": "",
            "axis/color": MAIN_FG_COLOR,
            "axis/title_font/size": 8,
            "axis/title_font/family": "default",
            "axis/title_font/bold": False,
            "axis/title_font/italic": False,
            "axis/ticks_font/size": 8,
            "axis/ticks_font/family": "default",
            "axis/ticks_font/bold": False,
            "axis/ticks_font/italic": False,
            # Default parameters for X/Y image axes
            "image_axis/title": "",
            "image_axis/unit": _("pixels"),
            "image_axis/color": MAIN_FG_COLOR,
            "image_axis/title_font/size": 8,
            "image_axis/title_font/family": "default",
            "image_axis/title_font/bold": False,
            "image_axis/title_font/italic": False,
            "image_axis/ticks_font/size": 8,
            "image_axis/ticks_font/family": "default",
            "image_axis/ticks_font/bold": False,
            "image_axis/ticks_font/italic": False,
            # Default parameters for color scale
            "color_axis/title": _("Intensity"),
            "color_axis/unit": _("lsb"),
            "color_axis/color": MAIN_FG_COLOR,
            "color_axis/title_font/size": 8,
            "color_axis/title_font/family": "default",
            "color_axis/title_font/bold": False,
            "color_axis/title_font/italic": False,
            "color_axis/ticks_font/size": 8,
            "color_axis/ticks_font/family": "default",
            "color_axis/ticks_font/bold": False,
            "color_axis/ticks_font/italic": False,
            "grid/background": MAIN_BG_COLOR,
            "grid/maj_xenabled": True,
            "grid/maj_yenabled": True,
            "grid/maj_line/color": "darkgray",
            "grid/maj_line/width": 1,
            "grid/maj_line/style": "DotLine",
            "grid/min_xenabled": True,
            "grid/min_yenabled": True,
            "grid/min_line/color": "#454545" if qth.is_dark_theme() else "#eaeaea",
            "grid/min_line/width": 1,
            "grid/min_line/style": "DotLine",
            "marker/curve/symbol/marker": "Rect",
            "marker/curve/symbol/edgecolor": "blue",
            "marker/curve/symbol/facecolor": "cyan",
            "marker/curve/symbol/alpha": 0.8,
            "marker/curve/symbol/size": 7,
            "marker/curve/text/font/size": 8,
            "marker/curve/text/font/family": "default",
            "marker/curve/text/font/bold": False,
            "marker/curve/text/font/italic": False,
            "marker/curve/text/textcolor": "black",
            "marker/curve/text/background_color": "#ffffff",
            "marker/curve/text/background_alpha": 0.8,
            "marker/curve/line/style": "SolidLine",
            "marker/curve/line/color": MAIN_FG_COLOR,
            "marker/curve/line/width": 0,
            "marker/curve/sel_symbol/marker": "Rect",
            "marker/curve/sel_symbol/edgecolor": "blue",
            "marker/curve/sel_symbol/facecolor": "cyan",
            "marker/curve/sel_symbol/alpha": 0.8,
            "marker/curve/sel_symbol/size": 7,
            "marker/curve/sel_text/font/size": 8,
            "marker/curve/sel_text/font/family": "default",
            "marker/curve/sel_text/font/bold": False,
            "marker/curve/sel_text/font/italic": False,
            "marker/curve/sel_text/textcolor": "black",
            "marker/curve/sel_text/background_color": "#ffffff",
            "marker/curve/sel_text/background_alpha": 0.8,
            "marker/curve/sel_line/style": "SolidLine",
            "marker/curve/sel_line/color": MAIN_FG_COLOR,
            "marker/curve/sel_line/width": 0,
            "marker/curve/markerstyle": "NoLine",
            "marker/curve/spacing": 7,
            "marker/cross/symbol/marker": "Cross",
            "marker/cross/symbol/edgecolor": MAIN_FG_COLOR,
            "marker/cross/symbol/facecolor": "red",
            "marker/cross/symbol/alpha": 1.0,
            "marker/cross/symbol/size": 8,
            "marker/cross/text/font/family": "default",
            "marker/cross/text/font/size": 8,
            "marker/cross/text/font/bold": False,
            "marker/cross/text/font/italic": False,
            "marker/cross/text/textcolor": "black",
            "marker/cross/text/background_color": "#ffffff",
            "marker/cross/text/background_alpha": 0.8,
            "marker/cross/line/style": "DashLine",
            "marker/cross/line/color": "yellow",
            "marker/cross/line/width": 1,
            "marker/cross/sel_symbol/marker": "Cross",
            "marker/cross/sel_symbol/edgecolor": MAIN_FG_COLOR,
            "marker/cross/sel_symbol/facecolor": "red",
            "marker/cross/sel_symbol/alpha": 1.0,
            "marker/cross/sel_symbol/size": 8,
            "marker/cross/sel_text/font/family": "default",
            "marker/cross/sel_text/font/size": 8,
            "marker/cross/sel_text/font/bold": False,
            "marker/cross/sel_text/font/italic": False,
            "marker/cross/sel_text/textcolor": "black",
            "marker/cross/sel_text/background_color": "#ffffff",
            "marker/cross/sel_text/background_alpha": 0.8,
            "marker/cross/sel_line/style": "DashLine",
            "marker/cross/sel_line/color": "yellow",
            "marker/cross/sel_line/width": 1,
            "marker/cross/markerstyle": "Cross",
            "marker/cross/spacing": 7,
            "marker/selectpoint/symbol/marker": "Rect",
            "marker/selectpoint/symbol/edgecolor": "blue",
            "marker/selectpoint/symbol/facecolor": "cyan",
            "marker/selectpoint/symbol/alpha": 0.8,
            "marker/selectpoint/symbol/size": 7,
            "marker/selectpoint/text/font/size": 8,
            "marker/selectpoint/text/font/family": "default",
            "marker/selectpoint/text/font/bold": False,
            "marker/selectpoint/text/font/italic": False,
            "marker/selectpoint/text/textcolor": "black",
            "marker/selectpoint/text/background_color": "#ffffff",
            "marker/selectpoint/text/background_alpha": 0.8,
            "marker/selectpoint/line/style": "SolidLine",
            "marker/selectpoint/line/color": MAIN_FG_COLOR,
            "marker/selectpoint/line/width": 0,
            "marker/selectpoint/sel_symbol/marker": "Rect",
            "marker/selectpoint/sel_symbol/edgecolor": "blue",
            "marker/selectpoint/sel_symbol/facecolor": "cyan",
            "marker/selectpoint/sel_symbol/alpha": 0.8,
            "marker/selectpoint/sel_symbol/size": 7,
            "marker/selectpoint/sel_text/font/size": 8,
            "marker/selectpoint/sel_text/font/family": "default",
            "marker/selectpoint/sel_text/font/bold": False,
            "marker/selectpoint/sel_text/font/italic": False,
            "marker/selectpoint/sel_text/textcolor": "black",
            "marker/selectpoint/sel_text/background_color": "#ffffff",
            "marker/selectpoint/sel_text/background_alpha": 0.8,
            "marker/selectpoint/sel_line/style": "SolidLine",
            "marker/selectpoint/sel_line/color": MAIN_FG_COLOR,
            "marker/selectpoint/sel_line/width": 0,
            "marker/selectpoint/markerstyle": "NoLine",
            "marker/selectpoint/spacing": 7,
            "marker/cursor/line/style": "SolidLine",
            "marker/cursor/line/color": "#ff9393",
            "marker/cursor/line/width": 1.0,
            "marker/cursor/symbol/marker": "Ellipse",
            "marker/cursor/symbol/size": 7,
            "marker/cursor/symbol/edgecolor": MAIN_BG_COLOR,
            "marker/cursor/symbol/facecolor": "#ff9393",
            "marker/cursor/symbol/alpha": 1.0,
            "marker/cursor/text/font/size": 8,
            "marker/cursor/text/font/family": "default",
            "marker/cursor/text/font/bold": False,
            "marker/cursor/text/font/italic": False,
            "marker/cursor/text/textcolor": "#ff9393",
            "marker/cursor/text/background_color": "#ffffff",
            "marker/cursor/text/background_alpha": 0.8,
            "marker/cursor/sel_line/style": "SolidLine",
            "marker/cursor/sel_line/color": "red",
            "marker/cursor/sel_line/width": 2.0,
            "marker/cursor/sel_symbol/marker": "Ellipse",
            "marker/cursor/sel_symbol/size": 9,
            "marker/cursor/sel_symbol/edgecolor": MAIN_BG_COLOR,
            "marker/cursor/sel_symbol/facecolor": "red",
            "marker/cursor/sel_symbol/alpha": 0.9,
            "marker/cursor/sel_text/font/size": 8,
            "marker/cursor/sel_text/font/family": "default",
            "marker/cursor/sel_text/font/bold": True,
            "marker/cursor/sel_text/font/italic": False,
            "marker/cursor/sel_text/textcolor": "red",
            "marker/cursor/sel_text/background_color": "#ffffff",
            "marker/cursor/sel_text/background_alpha": 0.8,
            "marker/cursor/markerstyle": "NoLine",
            "marker/cursor/spacing": 7,
            "shape/drag/line/style": "SolidLine",
            "shape/drag/line/color": "#ffff00",
            "shape/drag/line/width": 1,
            "shape/drag/fill/style": "SolidPattern",
            "shape/drag/fill/color": MAIN_BG_COLOR,
            "shape/drag/fill/alpha": 0.1,
            "shape/drag/symbol/marker": "Rect",
            "shape/drag/symbol/size": 3,
            "shape/drag/symbol/edgecolor": "#ffff00",
            "shape/drag/symbol/facecolor": "#ffff00",
            "shape/drag/symbol/alpha": 1.0,
            "shape/drag/sel_line/style": "SolidLine",
            "shape/drag/sel_line/color": "#00ff00",
            "shape/drag/sel_line/width": 1,
            "shape/drag/sel_fill/style": "SolidPattern",
            "shape/drag/sel_fill/color": MAIN_BG_COLOR,
            "shape/drag/sel_fill/alpha": 0.1,
            "shape/drag/sel_symbol/marker": "Rect",
            "shape/drag/sel_symbol/size": 9,
            "shape/drag/sel_symbol/edgecolor": "#00aa00",
            "shape/drag/sel_symbol/facecolor": "#00ff00",
            "shape/drag/sel_symbol/alpha": 0.7,
            "shape/imageborder/line/style": "NoPen",
            "shape/imageborder/line/color": "gray",
            "shape/imageborder/line/width": 0,
            "shape/imageborder/fill/style": "NoBrush",
            "shape/imageborder/fill/color": "black",
            "shape/imageborder/fill/alpha": 0.0,
            "shape/imageborder/fill/angle": 0.0,
            "shape/imageborder/fill/sx": 1.0,
            "shape/imageborder/fill/sy": 1.0,
            "shape/imageborder/symbol/marker": "NoSymbol",
            "shape/imageborder/symbol/size": 7,
            "shape/imageborder/symbol/edgecolor": "gray",
            "shape/imageborder/symbol/facecolor": MAIN_FG_COLOR,
            "shape/imageborder/symbol/alpha": 0.3,
            "shape/imageborder/sel_line/style": "SolidLine",
            "shape/imageborder/sel_line/color": "gray",
            "shape/imageborder/sel_line/width": 3,
            "shape/imageborder/sel_symbol/marker": "Rect",
            "shape/imageborder/sel_symbol/size": 7,
            "shape/imageborder/sel_symbol/edgecolor": "gray",
            "shape/imageborder/sel_symbol/facecolor": MAIN_FG_COLOR,
            "shape/imageborder/sel_symbol/alpha": 0.8,
            "shape/imageborder/sel_fill/style": "NoBrush",
            "shape/imageborder/sel_fill/color": "gray",
            "shape/imageborder/sel_fill/alpha": 0.5,
            "shape/imageborder/sel_fill/angle": 0.0,
            "shape/imageborder/sel_fill/sx": 1.0,
            "shape/imageborder/sel_fill/sy": 1.0,
            "shape/imagefilter/line/style": "SolidLine",
            "shape/imagefilter/line/color": "#ffff00",
            "shape/imagefilter/line/width": 1,
            "shape/imagefilter/sel_line/style": "SolidLine",
            "shape/imagefilter/sel_line/color": "#00ffff",
            "shape/imagefilter/sel_line/width": 2,
            "shape/imagefilter/fill/style": "NoBrush",
            "shape/imagefilter/fill/color": MAIN_BG_COLOR,
            "shape/imagefilter/fill/alpha": 0.0,
            "shape/imagefilter/sel_fill/style": "SolidPattern",
            "shape/imagefilter/sel_fill/color": MAIN_BG_COLOR,
            "shape/imagefilter/sel_fill/alpha": 0.2,
            "shape/imagefilter/symbol/marker": "Rect",
            "shape/imagefilter/symbol/size": 3,
            "shape/imagefilter/symbol/edgecolor": "#ffff00",
            "shape/imagefilter/symbol/facecolor": "#ffff00",
            "shape/imagefilter/symbol/alpha": 1.0,
            "shape/imagefilter/sel_symbol/marker": "Ellipse",
            "shape/imagefilter/sel_symbol/size": 7,
            "shape/imagefilter/sel_symbol/edgecolor": "#0000ff",
            "shape/imagefilter/sel_symbol/facecolor": "#00ffff",
            "shape/imagefilter/sel_symbol/alpha": 0.8,
            # Contour ------------------------------------------------------------------
            "shape/contour/line/style": "SolidLine",
            "shape/contour/line/color": "#000000",
            "shape/contour/line/width": 1,
            "shape/contour/fill/style": "NoBrush",
            "shape/contour/symbol/marker": "NoSymbol",
            "shape/contour/sel_line/style": "SolidLine",
            "shape/contour/sel_line/color": "#00ff00",
            "shape/contour/sel_line/width": 1,
            "shape/contour/sel_fill/style": "SolidPattern",
            "shape/contour/sel_fill/color": MAIN_BG_COLOR,
            "shape/contour/sel_fill/alpha": 0.1,
            "shape/contour/sel_symbol/marker": "NoSymbol",
            # RectZoom -----------------------------------------------------------------
            "shape/rectzoom/line/style": "SolidLine",
            "shape/rectzoom/line/color": "#bbbbbb",
            "shape/rectzoom/line/width": 2,
            # not used -- start
            "shape/rectzoom/sel_line/style": "SolidLine",
            "shape/rectzoom/sel_line/color": "green",
            "shape/rectzoom/sel_line/width": 2,
            # not used -- end
            "shape/rectzoom/fill/color": "yellow",
            "shape/rectzoom/fill/style": "SolidPattern",
            "shape/rectzoom/fill/alpha": 0.1,
            # not used -- start
            "shape/rectzoom/symbol/marker": "NoSymbol",
            "shape/rectzoom/symbol/size": 0,
            "shape/rectzoom/symbol/edgecolor": MAIN_FG_COLOR,
            "shape/rectzoom/symbol/facecolor": "yellow",
            "shape/rectzoom/symbol/alpha": 1.0,
            "shape/rectzoom/sel_symbol/marker": "Rect",
            "shape/rectzoom/sel_symbol/size": 5,
            "shape/rectzoom/sel_symbol/edgecolor": MAIN_FG_COLOR,
            "shape/rectzoom/sel_symbol/facecolor": "yellow",
            "shape/rectzoom/sel_symbol/alpha": 1.0,
            # not used -- end
            "shape/axes/border/line/style": "SolidLine",
            "shape/axes/border/line/color": "magenta",
            "shape/axes/border/line/width": 1,
            "shape/axes/border/sel_line/style": "SolidLine",
            "shape/axes/border/sel_line/color": "magenta",
            "shape/axes/border/sel_line/width": 2,
            "shape/axes/border/fill/style": "NoBrush",
            "shape/axes/border/fill/color": "magenta",
            "shape/axes/border/fill/alpha": 0.0,
            "shape/axes/border/sel_fill/style": "SolidPattern",
            "shape/axes/border/sel_fill/color": "magenta",
            "shape/axes/border/sel_fill/alpha": 0.3,
            "shape/axes/border/symbol/marker": "NoSymbol",
            "shape/axes/border/symbol/size": 0,
            "shape/axes/border/symbol/edgecolor": MAIN_FG_COLOR,
            "shape/axes/border/symbol/facecolor": "yellow",
            "shape/axes/border/symbol/alpha": 1.0,
            "shape/axes/border/sel_symbol/marker": "NoSymbol",
            "shape/axes/border/sel_symbol/size": 5,
            "shape/axes/border/sel_symbol/edgecolor": MAIN_FG_COLOR,
            "shape/axes/border/sel_symbol/facecolor": "yellow",
            "shape/axes/border/sel_symbol/alpha": 1.0,
            "shape/axes/arrow_size": 8,
            "shape/axes/arrow_angle": 30,
            "shape/axes/xarrow_pen/style": "SolidLine",
            "shape/axes/xarrow_pen/color": "red",
            "shape/axes/xarrow_pen/width": 1,
            "shape/axes/xarrow_brush/color": "red",
            "shape/axes/xarrow_brush/alpha": 0.2,
            "shape/axes/yarrow_pen/style": "SolidLine",
            "shape/axes/yarrow_pen/color": "green",
            "shape/axes/yarrow_pen/width": 1,
            "shape/axes/yarrow_brush/color": "green",
            "shape/axes/yarrow_brush/alpha": 0.2,
            "shape/image_stats/line/style": "DashLine",
            "shape/image_stats/line/color": "#ff88dc",
            "shape/image_stats/line/width": 1,
            "shape/image_stats/fill/style": "SolidPattern",
            "shape/image_stats/fill/color": MAIN_BG_COLOR,
            "shape/image_stats/fill/alpha": 0.1,
            "shape/image_stats/symbol/marker": "Rect",
            "shape/image_stats/symbol/size": 3,
            "shape/image_stats/symbol/edgecolor": "#ff55dc",
            "shape/image_stats/symbol/facecolor": "#ff88dc",
            "shape/image_stats/symbol/alpha": 0.6,
            "shape/image_stats/sel_line/style": "DashLine",
            "shape/image_stats/sel_line/color": "#ff00dc",
            "shape/image_stats/sel_line/width": 1,
            "shape/image_stats/sel_fill/style": "SolidPattern",
            "shape/image_stats/sel_fill/color": MAIN_BG_COLOR,
            "shape/image_stats/sel_fill/alpha": 0.1,
            "shape/image_stats/sel_symbol/marker": "Rect",
            "shape/image_stats/sel_symbol/size": 5,
            "shape/image_stats/sel_symbol/edgecolor": "#ff00dc",
            "shape/image_stats/sel_symbol/facecolor": "#ff88dc",
            "shape/image_stats/sel_symbol/alpha": 0.7,
            "shape/cross_section/line/style": "DotLine",
            "shape/cross_section/line/color": "#ff5555",
            "shape/cross_section/line/width": 1,
            "shape/cross_section/fill/style": "SolidPattern",
            "shape/cross_section/fill/color": MAIN_BG_COLOR,
            "shape/cross_section/fill/alpha": 0.1,
            "shape/cross_section/symbol/marker": "Cross",
            "shape/cross_section/symbol/size": 100,
            "shape/cross_section/symbol/edgecolor": "cyan",
            "shape/cross_section/symbol/facecolor": "cyan",
            "shape/cross_section/symbol/alpha": 0.6,
            "shape/cross_section/sel_line/style": "DotLine",
            "shape/cross_section/sel_line/color": "#ff0000",
            "shape/cross_section/sel_line/width": 1,
            "shape/cross_section/sel_fill/style": "SolidPattern",
            "shape/cross_section/sel_fill/color": MAIN_BG_COLOR,
            "shape/cross_section/sel_fill/alpha": 0.1,
            "shape/cross_section/sel_symbol/marker": "Cross",
            "shape/cross_section/sel_symbol/size": 10000,
            "shape/cross_section/sel_symbol/edgecolor": "cyan",
            "shape/cross_section/sel_symbol/facecolor": "cyan",
            "shape/cross_section/sel_symbol/alpha": 0.7,
            "shape/average_cross_section/line/style": "DotLine",
            "shape/average_cross_section/line/color": "#ff5555",
            "shape/average_cross_section/line/width": 1,
            "shape/average_cross_section/fill/style": "SolidPattern",
            "shape/average_cross_section/fill/color": MAIN_BG_COLOR,
            "shape/average_cross_section/fill/alpha": 0.1,
            "shape/average_cross_section/symbol/marker": "Diamond",
            "shape/average_cross_section/symbol/size": 7,
            "shape/average_cross_section/symbol/edgecolor": "#ff5555",
            "shape/average_cross_section/symbol/facecolor": "#ff5555",
            "shape/average_cross_section/symbol/alpha": 0.6,
            "shape/average_cross_section/sel_line/style": "DotLine",
            "shape/average_cross_section/sel_line/color": "#ff0000",
            "shape/average_cross_section/sel_line/width": 1,
            "shape/average_cross_section/sel_fill/style": "SolidPattern",
            "shape/average_cross_section/sel_fill/color": MAIN_BG_COLOR,
            "shape/average_cross_section/sel_fill/alpha": 0.1,
            "shape/average_cross_section/sel_symbol/marker": "Diamond",
            "shape/average_cross_section/sel_symbol/size": 9,
            "shape/average_cross_section/sel_symbol/edgecolor": "#aa0000",
            "shape/average_cross_section/sel_symbol/facecolor": "#ff0000",
            "shape/average_cross_section/sel_symbol/alpha": 0.7,
            "shape/mask/line/style": "DotLine",
            "shape/mask/line/color": "#5555ff",
            "shape/mask/line/width": 1,
            "shape/mask/fill/style": "SolidPattern",
            "shape/mask/fill/color": MAIN_BG_COLOR,
            "shape/mask/fill/alpha": 0.1,
            "shape/mask/symbol/marker": "Rect",
            "shape/mask/symbol/size": 7,
            "shape/mask/symbol/edgecolor": "#5555ff",
            "shape/mask/symbol/facecolor": "#5555ff",
            "shape/mask/symbol/alpha": 0.6,
            "shape/mask/sel_line/style": "DotLine",
            "shape/mask/sel_line/color": "#0000ff",
            "shape/mask/sel_line/width": 1,
            "shape/mask/sel_fill/style": "SolidPattern",
            "shape/mask/sel_fill/color": MAIN_BG_COLOR,
            "shape/mask/sel_fill/alpha": 0.1,
            "shape/mask/sel_symbol/marker": "Rect",
            "shape/mask/sel_symbol/size": 9,
            "shape/mask/sel_symbol/edgecolor": "#0000aa",
            "shape/mask/sel_symbol/facecolor": "#0000ff",
            "shape/mask/sel_symbol/alpha": 0.7,
            "shape/point/line/style": "SolidLine",
            "shape/point/line/color": "#ffff00",
            "shape/point/line/width": 1,
            "shape/point/sel_line/style": "SolidLine",
            "shape/point/sel_line/color": "#00ff00",
            "shape/point/sel_line/width": 1,
            "shape/point/fill/style": "NoBrush",
            "shape/point/sel_fill/style": "NoBrush",
            "shape/point/symbol/marker": "XCross",
            "shape/point/symbol/size": 9,
            "shape/point/symbol/edgecolor": "#ffff00",
            "shape/point/symbol/facecolor": "#ffff00",
            "shape/point/symbol/alpha": 1.0,
            "shape/point/sel_symbol/marker": "XCross",
            "shape/point/sel_symbol/size": 12,
            "shape/point/sel_symbol/edgecolor": "#00aa00",
            "shape/point/sel_symbol/facecolor": "#00ff00",
            "shape/point/sel_symbol/alpha": 0.7,
            "shape/segment/line/style": "SolidLine",
            "shape/segment/line/color": "#ffff00",
            "shape/segment/line/width": 1,
            "shape/segment/sel_line/style": "SolidLine",
            "shape/segment/sel_line/color": "#00ff00",
            "shape/segment/sel_line/width": 1,
            "shape/segment/fill/style": "NoBrush",
            "shape/segment/sel_fill/style": "NoBrush",
            "shape/segment/symbol/marker": "XCross",
            "shape/segment/symbol/size": 9,
            "shape/segment/symbol/edgecolor": "#ffff00",
            "shape/segment/symbol/facecolor": "#ffff00",
            "shape/segment/symbol/alpha": 1.0,
            "shape/segment/sel_symbol/marker": "XCross",
            "shape/segment/sel_symbol/size": 12,
            "shape/segment/sel_symbol/edgecolor": "#00aa00",
            "shape/segment/sel_symbol/facecolor": "#00ff00",
            "shape/segment/sel_symbol/alpha": 0.7,
            "shape/label/symbol/marker": "NoSymbol",
            "shape/label/symbol/size": 0,
            "shape/label/symbol/edgecolor": MAIN_BG_COLOR,
            "shape/label/symbol/facecolor": MAIN_BG_COLOR,
            "shape/label/border/style": "SolidLine",
            "shape/label/border/color": "#cbcbcb",
            "shape/label/border/width": 0,
            "shape/label/font/size": 8,
            "shape/label/font/family": "default",
            "shape/label/font/bold": False,
            "shape/label/font/italic": False,
            "shape/label/color": MAIN_BG_COLOR,
            "shape/label/bgcolor": MAIN_FG_COLOR,
            "shape/label/bgalpha": 0.25,
            "shape/label/abspos": False,
            "shape/label/move_anchor": True,
            "label/symbol/marker": "NoSymbol",
            "label/symbol/size": 0,
            "label/symbol/edgecolor": MAIN_BG_COLOR,
            "label/symbol/facecolor": MAIN_BG_COLOR,
            "label/border/style": "SolidLine",
            "label/border/color": "#cbcbcb",
            "label/border/width": 1,
            "label/font/size": 9,
            "label/font/family": "default",
            "label/font/bold": False,
            "label/font/italic": False,
            "label/color": MAIN_FG_COLOR,
            "label/bgcolor": MAIN_BG_COLOR,
            "label/bgalpha": 0.8,
            "label/anchor": "TL",
            "label/xc": 0,
            "label/yc": 0,
            "label/abspos": False,
            "label/absg": "TR",
            "label/xg": 0.0,
            "label/yg": 0.0,
            # info_label: used in builder.make.computation for example
            "info_label/symbol/marker": "NoSymbol",
            "info_label/symbol/size": 0,
            "info_label/symbol/edgecolor": MAIN_BG_COLOR,
            "info_label/symbol/facecolor": MAIN_BG_COLOR,
            "info_label/border/style": "SolidLine",
            "info_label/border/color": "#cbcbcb",
            "info_label/border/width": 1,
            "info_label/font/size": 9,
            "info_label/font/family": "default",
            "info_label/font/bold": False,
            "info_label/font/italic": False,
            "info_label/color": MAIN_FG_COLOR,
            "info_label/bgcolor": MAIN_BG_COLOR,
            "info_label/bgalpha": 0.8,
            "info_label/anchor": "TL",
            "info_label/xc": 0,
            "info_label/yc": 0,
            "info_label/abspos": True,
            "info_label/absg": "TR",
            "info_label/xg": 0.0,
            "info_label/yg": 0.0,
            "legend/border/style": "SolidLine",
            "legend/border/color": "#cbcbcb",
            "legend/border/width": 1,
            "legend/font/size": 8,
            "legend/font/family": "default",
            "legend/font/bold": False,
            "legend/font/italic": False,
            "legend/color": MAIN_FG_COLOR,
            "legend/bgcolor": MAIN_BG_COLOR,
            "legend/bgalpha": 0.8,
            "legend/anchor": "TR",
            "legend/xc": 0,
            "legend/yc": 0,
            "legend/abspos": True,
            "legend/absg": "TR",
            "legend/xg": 0.0,
            "legend/yg": 0.0,
        },
        "histogram": {
            "antialiasing": False,
            "title/font/size": 11,
            "title/font/bold": False,
            "label/font/size": 9,
            "label/font/bold": False,
            "curve/line/style": "NoPen",
            "curve/line/color": "green",
            "curve/line/width": 1,
            "curve/symbol/marker": "NoSymbol",
            "curve/symbol/size": 0,
            "curve/symbol/edgecolor": MAIN_BG_COLOR,
            "curve/symbol/facecolor": MAIN_FG_COLOR,
            "curve/symbol/alpha": 1.0,
            "curve/shade": 0.85,
            "curve/curvestyle": "Steps",
            "curve/label": "",
            "range/line/style": "SolidLine",
            "range/line/color": "#ff9393",
            "range/line/width": 1,
            "range/sel_line/style": "SolidLine",
            "range/sel_line/color": "red",
            "range/sel_line/width": 2,
            "range/fill": "red",
            "range/shade": 0.15,
            "range/symbol/marker": "Ellipse",
            "range/symbol/size": 7,
            "range/symbol/edgecolor": MAIN_BG_COLOR,
            "range/symbol/facecolor": "#ff9393",
            "range/symbol/alpha": 1.0,
            "range/sel_symbol/marker": "Ellipse",
            "range/sel_symbol/size": 9,
            "range/sel_symbol/edgecolor": MAIN_BG_COLOR,
            "range/sel_symbol/facecolor": "red",
            "range/sel_symbol/alpha": 0.9,
            "range/multi/color": "#806060",
            # Default parameters for plot axes
            "axis/title": "",
            "axis/color": MAIN_FG_COLOR,
            "axis/title_font/size": 7,
            "axis/title_font/family": "default",
            "axis/title_font/bold": False,
            "axis/title_font/italic": False,
            "axis/ticks_font/size": 7,
            "axis/ticks_font/family": "default",
            "axis/ticks_font/bold": False,
            "axis/ticks_font/italic": False,
            "grid/background": MAIN_BG_COLOR,
            "grid/maj_xenabled": True,
            "grid/maj_yenabled": True,
            "grid/maj_line/color": "darkgray",
            "grid/maj_line/width": 1,
            "grid/maj_line/style": "DotLine",
            "grid/min_xenabled": True,
            "grid/min_yenabled": False,
            "grid/min_line/color": "#eaeaea",
            "grid/min_line/width": 1,
            "grid/min_line/style": "DotLine",
            "marker/curve/symbol/marker": "Ellipse",
            "marker/curve/symbol/edgecolor": MAIN_FG_COLOR,
            "marker/curve/symbol/facecolor": "blue",
            "marker/curve/symbol/alpha": 0.8,
            "marker/curve/symbol/size": 8,
            "marker/curve/text/font/size": 8,
            "marker/curve/text/font/family": "default",
            "marker/curve/text/font/bold": False,
            "marker/curve/text/font/italic": False,
            "marker/curve/text/textcolor": "darkBlue",
            "marker/curve/text/background_color": "blue",
            "marker/curve/text/background_alpha": 0.2,
            "marker/curve/pen/style": "SolidLine",
            "marker/curve/pen/color": MAIN_FG_COLOR,
            "marker/curve/pen/width": 0,
            "marker/curve/markerstyle": "NoLine",
            "marker/cross/symbol/marker": "Cross",
            "marker/cross/symbol/edgecolor": MAIN_FG_COLOR,
            "marker/cross/symbol/facecolor": "red",
            "marker/cross/symbol/alpha": 1.0,
            "marker/cross/symbol/size": 8,
            "marker/cross/text/font/family": "default",
            "marker/cross/text/font/size": 8,
            "marker/cross/text/font/bold": False,
            "marker/cross/text/font/italic": False,
            "marker/cross/text/textcolor": "darkBlue",
            "marker/cross/text/background_color": "blue",
            "marker/cross/text/background_alpha": 0.2,
            "marker/cross/pen/style": "SolidLine",
            "marker/cross/pen/color": MAIN_FG_COLOR,
            "marker/cross/pen/width": 1,
            "marker/cross/markerstyle": "Cross",
        },
        "cross_section": {
            "antialiasing": False,
            "title/font/size": 11,
            "title/font/bold": False,
            "label/font/size": 9,
            "label/font/bold": False,
            "curve/line/style": "SolidLine",
            "curve/line/color": "blue",
            "curve/line/width": 1,
            "curve/symbol/marker": "NoSymbol",
            "curve/symbol/size": 0,
            "curve/symbol/edgecolor": MAIN_BG_COLOR,
            "curve/symbol/facecolor": MAIN_FG_COLOR,
            "curve/symbol/alpha": 1.0,
            "curve/shade": 0.0,
            "curve/label": "",
            "range/line/style": "SolidLine",
            "range/line/color": "#ff9393",
            "range/line/width": 1,
            "range/sel_line/style": "SolidLine",
            "range/sel_line/color": "red",
            "range/sel_line/width": 2,
            "range/fill": "red",
            "range/shade": 0.15,
            "range/symbol/marker": "NoSymbol",
            "range/symbol/size": 5,
            "range/symbol/edgecolor": MAIN_FG_COLOR,
            "range/symbol/facecolor": "yellow",
            "range/symbol/alpha": 1.0,
            "range/sel_symbol/marker": "Ellipse",
            "range/sel_symbol/size": 9,
            "range/sel_symbol/edgecolor": MAIN_BG_COLOR,
            "range/sel_symbol/facecolor": "red",
            "range/sel_symbol/alpha": 0.9,
            "range/multi/color": "#806060",
            # Default parameters for plot axes
            "axis/title": "",
            "axis/color": MAIN_FG_COLOR,
            "axis/title_font/size": 7,
            "axis/title_font/family": "default",
            "axis/title_font/bold": False,
            "axis/title_font/italic": False,
            "axis/ticks_font/size": 7,
            "axis/ticks_font/family": "default",
            "axis/ticks_font/bold": False,
            "axis/ticks_font/italic": False,
            "grid/background": MAIN_BG_COLOR,
            "grid/maj_xenabled": True,
            "grid/maj_yenabled": True,
            "grid/maj_line/color": "darkgray",
            "grid/maj_line/width": 1,
            "grid/maj_line/style": "DotLine",
            "grid/min_xenabled": False,
            "grid/min_yenabled": False,
            "grid/min_line/color": "#eaeaea",
            "grid/min_line/width": 1,
            "grid/min_line/style": "DotLine",
            "marker/curve/symbol/marker": "Ellipse",
            "marker/curve/symbol/edgecolor": MAIN_FG_COLOR,
            "marker/curve/symbol/facecolor": "blue",
            "marker/curve/symbol/alpha": 0.8,
            "marker/curve/symbol/size": 8,
            "marker/curve/text/font/size": 8,
            "marker/curve/text/font/family": "default",
            "marker/curve/text/font/bold": False,
            "marker/curve/text/font/italic": False,
            "marker/curve/text/textcolor": "darkBlue",
            "marker/curve/text/background_color": "blue",
            "marker/curve/text/background_alpha": 0.2,
            "marker/curve/pen/style": "SolidLine",
            "marker/curve/pen/color": MAIN_FG_COLOR,
            "marker/curve/pen/width": 0,
            "marker/curve/markerstyle": "NoLine",
            "marker/cross/symbol/marker": "Cross",
            "marker/cross/symbol/edgecolor": MAIN_FG_COLOR,
            "marker/cross/symbol/facecolor": "red",
            "marker/cross/symbol/alpha": 1.0,
            "marker/cross/symbol/size": 8,
            "marker/cross/text/font/family": "default",
            "marker/cross/text/font/size": 8,
            "marker/cross/text/font/bold": False,
            "marker/cross/text/font/italic": False,
            "marker/cross/text/textcolor": "darkBlue",
            "marker/cross/text/background_color": "blue",
            "marker/cross/text/background_alpha": 0.2,
            "marker/cross/pen/style": "SolidLine",
            "marker/cross/pen/color": MAIN_FG_COLOR,
            "marker/cross/pen/width": 1,
            "marker/cross/markerstyle": "Cross",
        },
        # colormaps options
        "colormaps": {
            "colormaps/default": "colormaps_default.json",
            "colormaps/custom": "colormaps_custom.json",
        },
    }


DEFAULTS = get_plotpy_defaults()

CONF = UserConfig(DEFAULTS)


def update_plotpy_defaults() -> None:
    """Update the defaults with the current configuration"""
    global DEFAULTS, MAIN_BG_COLOR, MAIN_FG_COLOR
    MAIN_FG_COLOR, MAIN_BG_COLOR = get_main_colors()
    DEFAULTS = get_plotpy_defaults()
    CONF.update_defaults(DEFAULTS)


def update_plotpy_color_mode() -> None:
    """Update the color mode for Plotpy"""
    update_plotpy_defaults()

    # Iterate over all `BasePlot` widgets by browsing the QApplication:
    from qtpy.QtWidgets import QApplication

    from plotpy.plot import BasePlot

    app = QApplication.instance()
    if app is not None:
        # Iterate over all widgets and children to find BasePlot instances:
        for widget in app.topLevelWidgets():
            for child in widget.findChildren(BasePlot):
                child.update_color_mode()


def set_plotpy_dark_mode(state: bool) -> None:
    """Set dark mode for Qt application

    Args:
        state (bool): True to enable dark mode
    """
    mode = "dark" if state else "light"
    warnings.warn(
        f"`set_plotpy_dark_mode` is deprecated and will be removed in a future "
        f"version. Use `set_plotpy_color_mode('{mode}')` instead.",
        DeprecationWarning,
    )
    set_plotpy_color_mode(mode)


def set_plotpy_color_mode(mode: Literal["light", "dark", "auto"] | None = None):
    """Set the color mode for Plotpy

    Args:
        mode: Color mode ('light', 'dark' or 'auto'). If 'auto', the system color mode
        is used. If None, the `QT_COLOR_MODE` environment variable is used.
    """
    qth.set_color_mode(mode)
    update_plotpy_color_mode()



# -*- coding: utf-8 -*-

from __future__ import annotations

from typing import TYPE_CHECKING

from guidata.dataset import DataSet, GetAttrProp, StringItem

from plotpy.config import _
from plotpy.styles.base import ItemParameters

if TYPE_CHECKING:
    from plotpy.items import PolygonMapItem


class PolygonMapParam(DataSet):
    """Dataset defining the parameters of a PolygonMapItem instance"""

    _multiselection = False
    label = StringItem(_("Title"), default="").set_prop(
        "display", hide=GetAttrProp("_multiselection")
    )

    def update_param(self, item: PolygonMapItem) -> None:
        """Updates the parameters using values from a given PolygonMapItem

        Args:
            item: reference PolygonMapItem instance
        """
        self.label = str(item.title().text())

    def update_item(self, item: PolygonMapItem) -> None:
        """Updates a given PolygonMapItem using the current parameters

        Args:
            item: instance of PolygonMapItem to update
        """
        plot = item.plot()
        if plot is not None:
            plot.blockSignals(True)  # Avoid unwanted calls of update_param
            # triggered by the setter methods below
        if not self._multiselection:
            # Non common parameters
            item.setTitle(self.label)
        if plot is not None:
            plot.blockSignals(False)


class PolygonParam_MS(PolygonMapParam):
    """Same as PolygonParam but for multiselection"""

    _multiselection = True


ItemParameters.register_multiselection(PolygonMapParam, PolygonParam_MS)



# -*- coding: utf-8 -*-

from __future__ import annotations

import warnings
from typing import TYPE_CHECKING

from guidata.dataset import (
    BeginGroup,
    BeginTabGroup,
    BoolItem,
    ChoiceItem,
    ColorItem,
    DataSet,
    EndGroup,
    EndTabGroup,
    FloatItem,
    GetAttrProp,
    IntItem,
    NotProp,
    Obj,
    StringItem,
    TextItem,
    update_dataset,
)
from qtpy import QtGui as QG

from plotpy.config import _
from plotpy.styles.base import FontItem, ItemParameters, LineStyleItem, SymbolItem

if TYPE_CHECKING:
    from plotpy.items import LabelItem, LegendBoxItem


class LabelParam(DataSet):
    """Parameters for a label item."""

    _multiselection = False
    _legend = False
    _no_contents = True
    label = StringItem(_("Title"), default="").set_prop(
        "display", hide=GetAttrProp("_multiselection")
    )

    _styles = BeginTabGroup("Styles")
    # -------------------------------------------------------------- Contents tab
    ___cont = BeginGroup(_("Contents")).set_prop(
        "display", icon="label.png", hide=GetAttrProp("_no_contents")
    )
    contents = TextItem("").set_prop("display", hide=GetAttrProp("_no_contents"))
    ___econt = EndGroup(_("Contents")).set_prop(
        "display", hide=GetAttrProp("_no_contents")
    )
    # ---------------------------------------------------------------- Symbol tab
    symbol = SymbolItem(_("Symbol")).set_prop(
        "display", icon="diamond.png", hide=GetAttrProp("_legend")
    )
    # ---------------------------------------------------------------- Border tab
    border = LineStyleItem(
        _("Border"), default=Obj(color="#cbcbcb"), help=_("set width to 0 to disable")
    ).set_prop("display", icon="dashdot.png")
    # ------------------------------------------------------------------ Text tab
    ___text = BeginGroup(_("Text")).set_prop("display", icon="font.png")
    font = FontItem(_("Text font"))
    color = ColorItem(_("Text color"), default="#000000")
    bgcolor = ColorItem(_("Background color"), default="#ffffff")
    bgalpha = FloatItem(_("Background transparency"), min=0.0, max=1.0, default=0.8)
    ___etext = EndGroup(_("Text"))
    # -------------------------------------------------------------- Position tab
    ___position = BeginGroup(_("Position")).set_prop("display", icon="move.png")
    _begin_anchor = BeginGroup(_("Position relative to anchor")).set_prop(
        "display", hide=GetAttrProp("_multiselection")
    )
    anchor = ChoiceItem(
        _("Corner"),
        [
            ("TL", _("Top left")),
            ("TR", _("Top right")),
            ("BL", _("Bottom left")),
            ("BR", _("Bottom right")),
            ("L", _("Left")),
            ("R", _("Right")),
            ("T", _("Top")),
            ("B", _("Bottom")),
            ("C", _("Center")),
        ],
        default="TL",
        help=_("Label position relative to anchor point"),
    ).set_prop("display", hide=GetAttrProp("_multiselection"))
    xc = IntItem(
        _("ΔX"),
        default=5,
        help=_("Horizontal offset (pixels) relative to anchor point"),
    ).set_prop("display", hide=GetAttrProp("_multiselection"))
    yc = (
        IntItem(
            _("ΔY"),
            default=5,
            help=_("Vertical offset (pixels) relative to anchor point"),
        )
        .set_pos(col=1)
        .set_prop("display", hide=GetAttrProp("_multiselection"))
    )
    _end_anchor = EndGroup(_("Anchor")).set_prop(
        "display", hide=GetAttrProp("_multiselection")
    )
    _begin_anchorpos = BeginGroup(_("Anchor position")).set_prop(
        "display", hide=GetAttrProp("_multiselection")
    )
    _abspos_prop = GetAttrProp("abspos")
    abspos = (
        BoolItem(text=_("Attach to canvas"), label=_("Anchor"), default=True)
        .set_prop("display", store=_abspos_prop)
        .set_prop("display", hide=GetAttrProp("_multiselection"))
    )
    xg = (
        FloatItem(_("X"), default=0.0, help=_("X-axis position in canvas coordinates"))
        .set_prop("display", active=NotProp(_abspos_prop))
        .set_prop("display", hide=GetAttrProp("_multiselection"))
    )
    yg = (
        FloatItem(_("Y"), default=0.0, help=_("Y-axis position in canvas coordinates"))
        .set_pos(col=1)
        .set_prop("display", active=NotProp(_abspos_prop))
        .set_prop("display", hide=GetAttrProp("_multiselection"))
    )
    move_anchor = (
        ChoiceItem(
            _("Interact"),
            (
                (True, _("moving object changes anchor position")),
                (False, _("moving object changes label position")),
            ),
            default=True,
        )
        .set_prop("display", active=NotProp(_abspos_prop))
        .set_prop("display", hide=GetAttrProp("_multiselection"))
    )
    absg = (
        ChoiceItem(
            _("Position"),
            [
                ("TL", _("Top left")),
                ("TR", _("Top right")),
                ("BL", _("Bottom left")),
                ("BR", _("Bottom right")),
                ("L", _("Left")),
                ("R", _("Right")),
                ("T", _("Top")),
                ("B", _("Bottom")),
                ("C", _("Center")),
            ],
            default="TL",
            help=_("Absolute position on canvas"),
        )
        .set_prop("display", active=_abspos_prop)
        .set_prop("display", hide=GetAttrProp("_multiselection"))
    )
    _end_anchorpos = EndGroup(_("Anchor position")).set_prop(
        "display", hide=GetAttrProp("_multiselection")
    )
    ___eposition = EndGroup(_("Position"))
    # ----------------------------------------------------------------------- End
    _endstyles = EndTabGroup("Styles")

    def update_param(self, item: LabelItem) -> None:
        """Update the parameters associated with the label item.

        Args:
            item: The label item from which to update the parameters.
        """
        # The following is necessary only for shape labels:
        # when shape is just created (and not yet moved), we need to update
        # these attributes
        update_dataset(self, item.labelparam)
        if self.abspos:
            self.absg = item.G
        else:
            self.xg, self.yg = item.G
        self.xc, self.yc = item.C
        self.label = item.title().text()

    def update_item(self, item: LabelItem) -> None:
        """Update the label item with the parameters.

        Args:
            item: The label item to update.
        """
        if not self._multiselection:
            if self.abspos:
                item.G = self.absg
            else:
                item.G = (self.xg, self.yg)
            item.C = self.xc, self.yc
            item.anchor = self.anchor
            item.setTitle(self.label)
        item.marker = self.symbol.build_symbol()
        item.border_pen = self.border.build_pen()
        item.set_text_style(self.font.build_font(), self.color)
        color = QG.QColor(self.bgcolor)
        color.setAlphaF(self.bgalpha)
        item.bg_brush = QG.QBrush(color)

    # TODO: remove this method in a future release
    def update_label(self, obj: LabelItem) -> None:
        """Update the label item with the parameters. This method is deprecated.

        Args:
            obj: The label item to update.
        """
        warnings.warn(
            "`LabelParam.update_label` method is deprecated and will be removed "
            "in a future release. Please use `update_item` instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        self.update_item(obj)


class LabelParam_MS(LabelParam):
    """Parameters for a label item in multiselection mode."""

    _multiselection = True


ItemParameters.register_multiselection(LabelParam, LabelParam_MS)


class LegendParam(LabelParam):
    """Parameters for a legend box item."""

    _legend = True
    label = StringItem(_("Title"), default="").set_prop("display", hide=True)

    def update_item(self, item: LegendBoxItem) -> None:
        """Update the legend item with the parameters.

        Args:
            item: The legend item to update.
        """
        super().update_item(item)
        if not self._multiselection:
            item.setTitle(self.get_title())


class LegendParam_MS(LegendParam):
    """Parameters for a legend box item in multiselection mode."""

    _multiselection = True


ItemParameters.register_multiselection(LegendParam, LegendParam_MS)


class LabelParamWithContents(LabelParam):
    """Parameters for a label item with contents.

    Args:
        title: The title of the label item.
        comment: The comment associated with the label item.
        icon: The icon associated with the label item.
    """

    _no_contents = False

    def __init__(
        self, title: str | None = None, comment: str | None = None, icon: str = ""
    ) -> None:
        self.plain_text = None
        super().__init__(title, comment, icon)

    def update_param(self, item: LabelItem) -> None:
        """Update the parameters associated with the label item.

        Args:
            item: The label item from which to update the parameters.
        """
        super().update_param(item)
        self.contents = self.plain_text = item.get_plain_text()

    def update_item(self, item: LabelItem) -> None:
        """Update the label item with the parameters.

        Args:
            item: The label item to update.
        """
        super().update_item(item)
        if self.plain_text is not None and self.contents != self.plain_text:
            text = self.contents.replace("\n", "<br>")
            item.set_text(text)


class LabelParamWithContents_MS(LabelParamWithContents):
    """Parameters for a label item with contents in multiselection mode."""

    _multiselection = True


ItemParameters.register_multiselection(
    LabelParamWithContents, LabelParamWithContents_MS
)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""Flip/rotate test"""

# guitest: show

from guidata.qthelpers import exec_dialog, qt_app_context
from qtpy import QtWidgets as QW

from plotpy.tests.widgets.test_rotatecrop import create_test_data, imshow
from plotpy.tools import RotationCenterTool
from plotpy.widgets.fliprotate import FlipRotateDialog, FlipRotateWidget


def widget_test(fname):
    """Test the flip/rotate widget"""
    array0, item = create_test_data(fname)
    widget = FlipRotateWidget(None, toolbar=True)
    widget.transform.set_item(item)
    widget.set_parameters(-90, True, False)
    widget.show()
    return widget


def dialog_test(fname):
    """Test the flip/rotate dialog with rotation point changeable"""
    array0, item = create_test_data(fname)
    dlg = FlipRotateDialog(None, toolbar=True)
    dlg.manager.add_tool(
        RotationCenterTool,
        rotation_center=False,
        rotation_point_move_with_shape=True,
        on_all_items=True,
    )
    dlg.transform.set_item(item)
    if exec_dialog(dlg) == QW.QDialog.Accepted:
        array1 = dlg.transform.get_result()
        dlg1 = imshow(array0, title="array0")
        dlg2 = imshow(array1, title="array1")
        return dlg, dlg1, dlg2


def test_flip_rotate():
    """Test the flip/rotate widget and dialog"""
    persist_list = []
    with qt_app_context(exec_loop=True):
        persist_list.append(widget_test("brain.png"))
    with qt_app_context(exec_loop=False):
        persist_list.append(dialog_test("brain.png"))


if __name__ == "__main__":
    test_flip_rotate()



# -*- coding: utf-8 -*-

from __future__ import annotations

from typing import TYPE_CHECKING

from guidata.dataset import BoolItem, ChoiceItem, ColorItem, DataSet, FloatItem, IntItem
from qtpy import QtGui as QG

from plotpy.config import _

if TYPE_CHECKING:
    from plotpy.items import ErrorBarCurveItem


class ErrorBarParam(DataSet):
    """Error bar style parameters for a curve."""

    mode = ChoiceItem(
        _("Display"),
        default=0,
        choices=[_("error bars with caps (x, y)"), _("error area (y)")],
        help=_(
            "Note: only y-axis error bars are shown in "
            "error area mode\n(width and cap parameters "
            "will also be ignored)"
        ),
    )
    color = ColorItem(_("Color"), default="darkred")
    alpha = FloatItem(
        _("Alpha"), default=0.9, min=0, max=1, help=_("Error bar transparency")
    )
    width = FloatItem(_("Width"), default=1.0, min=1)
    cap = IntItem(_("Cap"), default=4, min=0)
    ontop = BoolItem(_("set to foreground"), _("Visibility"), default=False)

    def update_param(self, item: ErrorBarCurveItem) -> None:
        """
        Update the parameters associated with the error bar item.

        Args:
            item: The error bar item from which to update the parameters.
        """
        color = item.errorPen.color()
        self.color = str(color.name())
        self.alpha = color.alphaF()
        self.width = item.errorPen.widthF()
        self.cap = item.errorCap
        self.ontop = item.errorOnTop

    def update_item(self, item: ErrorBarCurveItem) -> None:
        """
        Update the error bar item with the parameters.

        Args:
            item: The error bar item to update.
        """
        color = QG.QColor(self.color)
        color.setAlphaF(self.alpha)
        item.errorPen = QG.QPen(color, self.width)
        item.errorBrush = QG.QBrush(color)
        item.errorCap = self.cap
        item.errorOnTop = self.ontop



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""FontParam test"""

from guidata.qthelpers import qt_app_context

from plotpy.styles.base import FontParam


def test_fontparam():
    with qt_app_context(exec_loop=True):
        fp = FontParam()
        fp.edit()
        fp.edit()


if __name__ == "__main__":
    test_fontparam()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""Plot computations test"""

# guitest: show

import numpy as np
import scipy.integrate as spt
from guidata.qthelpers import qt_app_context

from plotpy.builder import make
from plotpy.tests import vistools as ptv


def test_computations():
    """Test computations"""
    x = np.linspace(-10, 10, 1000)
    y = np.sin(np.sin(np.sin(x)))
    with qt_app_context(exec_loop=True):
        curve = make.curve(x, y, "ab", "b")
        range = make.range(-2, 2)
        disp0 = make.range_info_label(
            range, "BR", "x = %.1f ± %.1f cm", title="Range infos"
        )

        disp1 = make.computation(
            range, "BL", "trapz=%g", curve, lambda x, y: spt.trapezoid(y, x)
        )

        disp2 = make.computations(
            range,
            "TL",
            [
                (curve, "min=%.5f", lambda x, y: y.min()),
                (curve, "max=%.5f", lambda x, y: y.max()),
                (curve, "avg=%.5f", lambda x, y: y.mean()),
            ],
        )
        legend = make.legend("TR")
        _win = ptv.show_items(
            wintitle="Plot computations",
            items=[curve, range, disp0, disp1, disp2, legend],
            plot_type="curve",
        )


if __name__ == "__main__":
    test_computations()



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import pytest
from guidata.qthelpers import exec_dialog, execenv, qt_app_context

from plotpy.interfaces.items import IBasePlotItem, IShapeItemType
from plotpy.tests import vistools as ptv
from plotpy.tests.features.test_auto_curve_image import make_curve_image_legend
from plotpy.tools import (
    AnnotatedCircleTool,
    AnnotatedEllipseTool,
    AnnotatedObliqueRectangleTool,
    AnnotatedPointTool,
    AnnotatedPolygonTool,
    AnnotatedRectangleTool,
    AnnotatedSegmentTool,
    AverageCrossSectionTool,
    CircleTool,
    CrossSectionTool,
    EllipseTool,
    ImageStatsTool,
    InteractiveTool,
    LabelTool,
    ObliqueRectangleTool,
    PointTool,
    RectangleTool,
    SegmentTool,
    SelectTool,
    SnapshotTool,
)

if TYPE_CHECKING:
    from plotpy.items.image.base import BaseImageItem
    from plotpy.plot.plotwidget import PlotWindow

from plotpy.tests.unit.utils import drag_mouse, undo_redo

TOOLS: tuple[type[InteractiveTool], ...] = (
    AnnotatedCircleTool,
    AnnotatedEllipseTool,
    AnnotatedObliqueRectangleTool,
    AnnotatedPointTool,
    AnnotatedRectangleTool,
    AnnotatedSegmentTool,
    AnnotatedPolygonTool,
    AverageCrossSectionTool,
    CrossSectionTool,
    SnapshotTool,
    ImageStatsTool,
    CircleTool,
    EllipseTool,
    ObliqueRectangleTool,
    PointTool,
    RectangleTool,
    SegmentTool,
    LabelTool,
)


def create_window(tool_classes: tuple[type[InteractiveTool], ...]) -> PlotWindow:
    """Create a window with the given tools. The plot contains a curve, an image and a
     legend.

    Args:
        tool_classes: tools to add to the window.

    Returns:
        PlotWindow: The window containing the tools.
    """
    items = make_curve_image_legend()
    img: BaseImageItem = items[0]
    img.set_rotatable(True)
    img.set_movable(True)
    win = ptv.show_items(
        items, wintitle="Unit tests for RectangularActionTools", auto_tools=True
    )

    for toolklass in tool_classes:
        _ = win.manager.add_tool(toolklass)

    assert len(tool_classes) > 0

    return win


def _test_annotation_tools(tool_classes: tuple[type[InteractiveTool], ...]):
    """Generic test for annotation tool. Simulates a mouse drag on the plot and checks
    that the tool is activated and deactivated correctly."""
    with qt_app_context(exec_loop=False):
        win = create_window(tool_classes)
        win.show()
        plot = win.manager.get_plot()
        default_tool = win.manager.get_default_tool()
        for tool_class in tool_classes:
            tool = win.manager.get_tool(tool_class)
            assert tool is not None
            tool.activate()
            x_path = np.linspace(0, 0.5, 100)
            y_path = np.linspace(0, 0.5, 100)
            with execenv.context(accept_dialogs=True):
                drag_mouse(win, x_path, y_path)
            if hasattr(tool_class, "SWITCH_TO_DEFAULT_TOOL"):
                assert win.manager.get_default_tool() == default_tool
        plot.unselect_all()
        shape_items: list[IBasePlotItem] = plot.get_items(item_type=IShapeItemType)
        plot.select_some_items(shape_items)
        select_tool = win.manager.get_tool(SelectTool)
        assert select_tool is not None
        select_tool.activate()

        drag_mouse(win, np.linspace(0.2, 0.5, 10), np.linspace(0.2, 0.5, 10))

        undo_redo(win)

        exec_dialog(win)


@pytest.mark.parametrize("tool", TOOLS)
def test_tool(tool: type[InteractiveTool]) -> None:
    """Test a single tool."""
    _test_annotation_tools((tool,))


def _test_all_tools():
    """Test all tools."""
    _test_annotation_tools(TOOLS)


if __name__ == "__main__":
    _test_all_tools()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""
plotpy test package
===================
"""

import os.path as osp

from guidata.configtools import get_module_data_path

import plotpy  # noqa: F401

TESTDATAPATH = get_module_data_path("plotpy", osp.join("tests", "data"))


def get_path(filename: str) -> str:
    """Return absolute path of test file"""
    return osp.join(TESTDATAPATH, filename)


def run() -> None:
    """Run plotpy test launcher"""
    from guidata.guitest import run_testlauncher

    import plotpy.config  # load icons

    run_testlauncher(plotpy)


if __name__ == "__main__":
    run()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

# -*- coding: utf-8 -*-
"""
Testing plotpy QtDesigner plugins

These plugins provide PlotWidget objects
embedding in GUI layouts directly from QtDesigner.
"""

# guitest: show

import os

import pytest
from guidata.qthelpers import qt_app_context

from plotpy.builder import make
from plotpy.tests import data as ptd

try:
    from plotpy.widgets.qtdesigner import loadui
except ImportError:
    # PySide6 known to fail
    pytest.skip(
        "PySide6 does not support QPyDesignerCustomWidgetPlugin",
        allow_module_level=True,
    )

FormClass = loadui(os.path.splitext(__file__)[0] + ".ui")


class WindowTest(FormClass):
    def __init__(self, image_data):
        super().__init__()
        plot = self.imagewidget.plot
        plot.add_item(make.image(image_data))
        self.setWindowTitle("QtDesigner plugins example")


def test_qtdesigner():
    with qt_app_context(exec_loop=True):
        form = WindowTest(ptd.gen_image4(200, 200))
        form.show()


if __name__ == "__main__":
    test_qtdesigner()



# -*- coding: utf-8 -*-

from __future__ import annotations

from typing import TYPE_CHECKING

from guidata.configtools import get_icon

from plotpy.config import _
from plotpy.constants import ID_LCS, ID_OCS, ID_XCS, ID_YCS
from plotpy.interfaces import IImageItemType
from plotpy.items import (
    AnnotatedObliqueRectangle,
    AnnotatedPoint,
    AnnotatedRectangle,
    AnnotatedSegment,
)
from plotpy.styles import AnnotationParam
from plotpy.tools.base import PanelTool
from plotpy.tools.image import update_image_tool_status
from plotpy.tools.shape import RectangularShapeTool

if TYPE_CHECKING:
    from plotpy.plot import BasePlot


class CrossSectionTool(RectangularShapeTool):
    SWITCH_TO_DEFAULT_TOOL = True
    TITLE = _("Cross section")
    ICON = "csection.png"
    SHAPE_STYLE_KEY = "shape/cross_section"
    SHAPE_TITLE = TITLE
    PANEL_IDS = (ID_XCS, ID_YCS)

    def update_status(self, plot: BasePlot) -> None:
        """Update status of the tool"""
        if update_image_tool_status(self, plot):
            item = plot.get_selected_items(item_type=IImageItemType)
            self.action.setEnabled(len(item) > 0)

    def create_shape(self) -> AnnotatedPoint:
        """Create shape for the tool"""
        return AnnotatedPoint(0, 0), 0, 0

    def setup_shape(self, shape: AnnotatedPoint) -> None:
        """Set up shape for the tool"""
        self.setup_shape_appearance(shape)
        super().setup_shape(shape)
        self.register_shape(shape)

    def setup_shape_appearance(self, shape: AnnotatedPoint) -> None:
        """Set up shape appearance"""
        self.set_shape_style(shape)
        param = shape.annotationparam
        param.title = self.SHAPE_TITLE
        #        param.show_computations = False
        param.update_item(shape)

    def register_shape(self, shape: AnnotatedPoint) -> None:
        """Register shape"""
        plot = shape.plot()
        if plot is not None:
            plot.unselect_all()
            plot.set_active_item(shape)
        for panel_id in self.PANEL_IDS:
            panel = self.manager.get_panel(panel_id)
            if panel is not None:
                panel.register_shape(shape)

    def activate(self) -> None:
        """Activate tool"""
        super().activate()
        for panel_id in self.PANEL_IDS:
            panel = self.manager.get_panel(panel_id)
            panel.setVisible(True)
            shape = self.get_last_final_shape()
            if shape is not None:
                panel.update_plot(shape)

    def handle_final_shape(self, shape: AnnotatedPoint) -> None:
        """Handle final shape"""
        super().handle_final_shape(shape)
        self.register_shape(shape)


class AverageCrossSectionTool(CrossSectionTool):
    SWITCH_TO_DEFAULT_TOOL = True
    TITLE = _("Average cross section")
    ICON = "csection_a.png"
    SHAPE_STYLE_KEY = "shape/average_cross_section"
    SHAPE_TITLE = TITLE

    def create_shape(self) -> AnnotatedRectangle:
        """Create shape for the tool"""
        return AnnotatedRectangle(0, 0, 1, 1), 0, 2


class LineCrossSectionTool(CrossSectionTool):
    SWITCH_TO_DEFAULT_TOOL = True
    TITLE = _("Line cross section")
    ICON = "csection_line.png"
    SHAPE_STYLE_KEY = "shape/average_cross_section"
    SHAPE_TITLE = TITLE
    PANEL_IDS = (ID_LCS,)

    def create_shape(self) -> AnnotatedSegment:
        """Create shape for the tool"""
        param = AnnotationParam.create(show_label=False, show_computations=False)
        return AnnotatedSegment(0, 0, 1, 1, param), 0, 1


class ObliqueCrossSectionTool(CrossSectionTool):
    SWITCH_TO_DEFAULT_TOOL = True
    TITLE = _("Oblique averaged cross section")
    ICON = "csection_oblique.png"
    SHAPE_STYLE_KEY = "shape/average_cross_section"
    SHAPE_TITLE = TITLE
    PANEL_IDS = (ID_OCS,)

    def create_shape(self) -> AnnotatedObliqueRectangle:
        """Create shape for the tool"""
        annotation = AnnotatedObliqueRectangle(0, 0, 1, 0, 1, 1, 0, 1)
        self.set_shape_style(annotation)
        annotation.setIcon(get_icon(self.ICON))
        return annotation, 0, 2


class XCSPanelTool(PanelTool):
    panel_name = _("X-axis cross section")
    panel_id = ID_XCS


class YCSPanelTool(PanelTool):
    panel_name = _("Y-axis cross section")
    panel_id = ID_YCS


class OCSPanelTool(PanelTool):
    panel_name = _("Oblique averaged cross section")
    panel_id = ID_OCS


class LCSPanelTool(PanelTool):
    panel_name = _("Line cross section")
    panel_id = ID_LCS



from __future__ import annotations

from typing import Callable

import numpy as np
from guidata.qthelpers import exec_dialog, qt_app_context
from numpy import linspace

from plotpy.interfaces.items import ICurveItemType
from plotpy.tests.unit.utils import create_window, drag_mouse
from plotpy.tools import RectZoomTool


def zoom(
    x_path: np.ndarray, y_path: np.ndarray, compare: Callable[[float, float], bool]
):
    """Zoom in or out on a plot by sumilating a mouse drag along a defined path.

    Args:
        x_path: The x relative plot coordinates of the path to simulate the mouse
         drag along.
        y_path: The y relative plot coordinates of the path to simulate the mouse
         drag along.
        compare: Comparison function to compare the initial and final axis limits to
         check if the zoom worked as expected.
    """
    with qt_app_context(exec_loop=False):
        win, _tool = create_window(RectZoomTool, active_item_type=ICurveItemType)
        plot = win.manager.get_plot()
        initial_axis_limits = [
            plot.get_axis_limits(axis) for axis in plot.get_active_axes()
        ]

        drag_mouse(win, x_path, y_path)

        final_axis_limits = [
            plot.get_axis_limits(axis) for axis in plot.get_active_axes()
        ]

        print(initial_axis_limits)
        print(final_axis_limits)

        assert all(
            map(
                lambda og, final: compare(og[1], final[1]),
                initial_axis_limits,
                final_axis_limits,
            )
        )
        exec_dialog(win)


def test_rect_zoom_tool():
    """Test the rectangular zoom tool."""
    x_path = linspace(0, 0.5, 100)
    y_path = linspace(0, 0.4, 100)
    zoom(x_path, y_path, lambda og, final: final < og)


def test_rect_unzoom_tool():
    """Test the rectangular zoom tool to unzoom."""
    x_path = linspace(-0.5, 1.5, 100)
    y_path = linspace(-0.5, 1.5, 100)
    zoom(x_path, y_path, lambda og, final: abs(final) > abs(og))


if __name__ == "__main__":
    test_rect_zoom_tool()
    test_rect_unzoom_tool()



class IPanel:
    """Interface for panels controlled by PlotManager"""

    @staticmethod
    def __inherits__():
        # Avoid circular import
        # pylint: disable=import-outside-toplevel
        from plotpy.panels import PanelWidget

        return PanelWidget

    def register_panel(self, manager):
        """Register panel to plot manager"""
        pass

    def configure_panel(self):
        """Configure panel"""
        pass



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""
Dot array example
=================

Example showing how to create a custom item (drawing dots of variable size)
and integrate the associated `guidata` dataset (GUI-based form) to edit its
parameters (directly into the same window as the plot itself, *and* within
the custom item parameters: right-click on the selectable item to open the
associated dialog box).
"""

# guitest: show

from __future__ import annotations

from typing import TYPE_CHECKING

import guidata.dataset as gds
import guidata.dataset.qtwidgets as gdq
import numpy as np
from guidata.configtools import get_image_file_path
from guidata.qthelpers import qt_app_context
from qtpy import QtCore as QC
from qtpy import QtGui as QG
from qtpy import QtWidgets as QW

import plotpy.config  # Loading icons  # noqa: F401
from plotpy.interfaces import IImageItemType
from plotpy.items import RawImageItem
from plotpy.items.curve.errorbar import vmap
from plotpy.plot import PlotDialog, PlotOptions
from plotpy.styles import RawImageParam
from plotpy.tools import CopyToClipboardTool, HelpTool, PrintTool, SaveAsTool

if TYPE_CHECKING:
    from plotpy.interfaces import IItemType


class DotArrayParam(gds.DataSet):
    """Dot array"""

    def _update_cb(self, *args):
        """Update callback, to be overriden"""
        pass

    g1 = gds.BeginGroup("Size of the area")
    dim_h = gds.FloatItem("Width", default=20, min=0, unit="mm")
    dim_v = gds.FloatItem("Height", default=20, min=0, unit="mm")
    _g1 = gds.EndGroup("Size of the area")

    g2 = gds.BeginGroup("Grid pattern properties")
    step_x = gds.FloatItem("Step in X-axis", default=1, min=1, unit="mm")
    step_y = gds.FloatItem("Step in Y-axis", default=1, min=1, unit="mm")
    size = gds.FloatItem("Dot size", default=0.2, min=0, max=2, slider=True, unit="mm")
    color = gds.ColorItem("Dot color", default="red")
    _g2 = gds.EndGroup("Grid pattern properties")

    def update_item(self, obj):
        """Update item from parameters"""
        self._update_cb()

    def update_param(self, obj):
        """Update parameters from object"""
        pass


class DotArrayRawImageParam(RawImageParam, DotArrayParam):
    pass


class DotArrayItem(RawImageItem):
    """Dot array item"""

    def __init__(self, param=None):
        super().__init__(np.zeros((1, 1)), param)
        self.update_border()

    def boundingRect(self):
        """Reimplemented to return the bounding rectangle of the item"""
        param = self.param
        if param is not None:
            return QC.QRectF(
                QC.QPointF(-0.5 * param.size, -0.5 * param.size),
                QC.QPointF(
                    param.dim_h + 0.5 * param.size, param.dim_v + 0.5 * param.size
                ),
            )

    def types(self) -> tuple[type[IItemType], ...]:
        """Returns a group or category for this item.
        This should be a tuple of class objects inheriting from IItemType

        Returns:
            tuple: Tuple of class objects inheriting from IItemType
        """
        return (IImageItemType,)

    def draw_image(self, painter, canvasRect, srcRect, dstRect, xMap, yMap):
        """Draw image"""
        painter.setRenderHint(QG.QPainter.Antialiasing, True)
        param = self.param
        xcoords = vmap(xMap, np.arange(0, param.dim_h + 1, param.step_x))
        ycoords = vmap(yMap, np.arange(0, param.dim_v + 1, param.step_y))
        rx = 0.5 * param.size * xMap.pDist() / xMap.sDist()
        ry = 0.5 * param.size * yMap.pDist() / yMap.sDist()
        color = QG.QColor(param.color)
        painter.setPen(QG.QPen(color))
        painter.setBrush(QG.QBrush(color))
        for xc in xcoords:
            for yc in ycoords:
                painter.drawEllipse(QC.QPointF(xc, yc), rx, ry)


class CustomHelpTool(HelpTool):
    """Custom help tool"""

    def activate_command(self, plot, checked):
        """Activate command"""
        QW.QMessageBox.information(
            plot,
            "Help",
            """**to be customized**
Keyboard/mouse shortcuts:
  - single left-click: item (curve, image, ...) selection
  - single right-click: context-menu relative to selected item
  - shift: on-active-curve (or image) cursor
  - alt: free cursor
  - left-click + mouse move: move item (when available)
  - middle-click + mouse move: pan
  - right-click + mouse move: zoom""",
        )


class DotArrayDialog(PlotDialog):
    """Dot array dialog"""

    def __init__(self):
        self.item = None
        self.stamp_gbox = None
        super().__init__(
            title="Dot array example",
            options=PlotOptions(title="Main plot", type="image"),
            toolbar=True,
            edit=True,
        )
        self.resize(900, 600)

    def register_tools(self):
        """Register tools"""
        manager = self.plot_widget.manager
        manager.register_standard_tools()
        manager.add_separator_tool()
        manager.add_tool(SaveAsTool)
        manager.add_tool(CopyToClipboardTool)
        manager.add_tool(PrintTool)
        manager.add_tool(CustomHelpTool)
        manager.activate_default_tool()
        plot = manager.get_plot()
        plot.enableAxis(plot.yRight, False)
        plot.set_aspect_ratio(lock=True)

    def populate_plot_layout(self):
        """Populate the plot layout

        Reimplements the method from PlotDialog"""
        self.add_widget(self.plot_widget, row=0, column=0, rowspan=3, columnspan=1)
        logo_path = get_image_file_path("plotpy.svg")
        logo = QW.QLabel()
        logo.setPixmap(QG.QPixmap(logo_path))
        logo.setAlignment(QC.Qt.AlignCenter)
        self.add_widget(logo, 1, 1)
        logo_txt = QW.QLabel("Powered by <b>plotpy</b>")
        logo_txt.setAlignment(QC.Qt.AlignHCenter | QC.Qt.AlignTop)
        self.add_widget(logo_txt, 2, 1)
        self.stamp_gbox = gdq.DataSetEditGroupBox("Dots", DotArrayParam)
        self.stamp_gbox.SIG_APPLY_BUTTON_CLICKED.connect(self.apply_params)
        self.add_widget(self.stamp_gbox, 0, 1)

    def show_data(self, param):
        """Show data"""
        plot = self.plot_widget.plot
        if self.item is None:
            itemparam = DotArrayRawImageParam()
            gds.update_dataset(itemparam, param)
            param._update_cb = lambda: self.stamp_gbox.get()
            self.item = DotArrayItem(itemparam)
            plot.add_item(self.item)
        else:
            gds.update_dataset(self.item.param, param)
            self.item.update_border()
        plot.do_autoscale()

    def apply_params(self):
        """Apply parameters"""
        param = self.stamp_gbox.dataset
        self.show_data(param)


def test_dot_array():
    """Test dot array dialog"""
    with qt_app_context(exec_loop=True):
        dlg = DotArrayDialog()
        dlg.apply_params()
        dlg.show()


if __name__ == "__main__":
    test_dot_array()



# -*- coding: utf-8 -*-
#
# This file is part of CodraFT Project
# https://codra-ingenierie-informatique.github.io/CodraFT/
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
Unit test for plot items <--> JSON serialization/deserialization

How to save/restore items to/from a JSON string?

    # Plot items --> JSON:
    writer = JSONWriter(None)
    save_items(writer, items)
    text = writer.get_json()

    # JSON --> Plot items:
    items = load_items(JSONReader(text))

"""

# guitest: show

# WARNING:
# This script requires read/write permissions on current directory

from __future__ import annotations

from guidata.io import JSONReader, JSONWriter

from plotpy.tests.features.test_loadsaveitems_pickle import IOTest


class JSONTest(IOTest):
    """Class for JSON I/O testing"""

    FNAME = "loadsavecanvas.json"

    def restore_items(self) -> None:
        """Restore plot items"""
        self.plot.deserialize(JSONReader(self.FNAME))

    def save_items(self) -> None:
        """Save plot items"""
        writer = JSONWriter(self.FNAME)
        self.plot.serialize(writer)
        writer.save()


def test_loadsaveitems_json():
    """Test load/save items from/to JSON file"""
    test = JSONTest("Load/save items from/to JSON file")
    test.run()


if __name__ == "__main__":
    test_loadsaveitems_json()



# -*- coding: utf-8 -*-

from __future__ import annotations

import warnings
from typing import TYPE_CHECKING

from guidata.dataset import (
    BeginGroup,
    ChoiceItem,
    ColorItem,
    DataSet,
    EndGroup,
    FloatItem,
    ObjectItem,
    StringItem,
)
from guidata.dataset.qtitemwidgets import DataSetWidget
from guidata.dataset.qtwidgets import DataSetEditLayout
from qwt import QwtPlot

from plotpy.config import _
from plotpy.constants import X_BOTTOM, Y_LEFT, Y_RIGHT
from plotpy.styles.base import FontItem

if TYPE_CHECKING:
    from qwt import QwtPlotItem, QwtScaleDiv

    from plotpy.items import BaseImageItem
    from plotpy.plot import BasePlot


class AxeStyleParam(DataSet):
    """Style parameters for an axis."""

    title = StringItem(_("Title"), default="")
    unit = StringItem(_("Unit"), default="")
    color = ColorItem(_("Color"), default="black").set_pos(col=1)
    title_font = FontItem(_("Title font"))
    ticks_font = FontItem(_("Values font"))


class AxisParam(DataSet):
    """Scale parameters for an axis."""

    scale = ChoiceItem(
        _("Scale"), [("lin", _("linear")), ("log", _("logarithmic"))], default="lin"
    )
    vmin = FloatItem("Min", help=_("Lower axis limit"))
    vmax = FloatItem("Max", help=_("Upper axis limit"))

    def update_param(self, plot: BasePlot, axis_id: int) -> None:
        """
        Update the parameters of the axis.

        Args:
            plot: The plot from which to update the parameters.
            axis_id: The axis ID to update.
        """
        self.scale = plot.get_axis_scale(axis_id)
        axis: QwtScaleDiv = plot.axisScaleDiv(axis_id)
        self.vmin = axis.lowerBound()
        self.vmax = axis.upperBound()

    def update_axis(self, plot: BasePlot, axis_id: int) -> None:
        """
        Update the axis with the parameters.

        Args:
            plot: The plot to update.
            axis_id: The axis ID to update.
        """
        plot.enableAxis(axis_id, True)
        plot.set_axis_scale(axis_id, self.scale, autoscale=False)
        plot.setAxisScale(axis_id, self.vmin, self.vmax)
        plot.disable_unused_axes()
        plot.SIG_AXIS_PARAMETERS_CHANGED.emit(axis_id)


class AxisItemWidget(DataSetWidget):
    """Widget for the axis item."""

    klass = AxisParam


class AxisItem(ObjectItem):
    """Item for an axis."""

    klass = AxisParam


DataSetEditLayout.register(AxisItem, AxisItemWidget)


class AxesParam(DataSet):
    """Parameters for the axes of a plot."""

    xaxis_id = ChoiceItem(
        _("X-axis position"),
        [(QwtPlot.xBottom, _("bottom")), (QwtPlot.xTop, _("top"))],
        default=QwtPlot.xBottom,
    )
    xaxis = AxisItem(_("X Axis"))
    yaxis_id = ChoiceItem(
        _("Y-axis position"),
        [(QwtPlot.yLeft, _("left")), (QwtPlot.yRight, _("right"))],
        default=QwtPlot.yLeft,
    )
    yaxis = AxisItem(_("Y Axis"))

    def update_param(self, item: QwtPlotItem) -> None:
        """
        Update the parameters of the axes.

        Args:
            item: The plot item from which to update the parameters.
        """
        plot: BasePlot = item.plot()
        self.xaxis: AxisParam
        self.yaxis: AxisParam
        self.xaxis_id = item.xAxis()
        self.xaxis.update_param(plot, self.xaxis_id)
        self.yaxis_id = item.yAxis()
        self.yaxis.update_param(plot, self.yaxis_id)

    def update_item(self, item: QwtPlotItem) -> None:
        """
        Update the axes with the parameters.

        Args:
            item: The plot item to update.
        """
        plot: BasePlot = item.plot()
        self.xaxis: AxisParam
        self.yaxis: AxisParam
        plot.grid.setAxes(self.xaxis_id, self.yaxis_id)
        item.setXAxis(self.xaxis_id)
        self.xaxis.update_axis(plot, self.xaxis_id)
        item.setYAxis(self.yaxis_id)
        self.yaxis.update_axis(plot, self.yaxis_id)

    # TODO: remove this method in a future release
    def update_axes(self, obj: QwtPlotItem) -> None:
        """Update object from parameters. Deprecated, use update_item instead.

        Args:
            item: The plot item to update.
        """
        warnings.warn(
            "`AxesParam.update_axes` method is deprecated and will be removed "
            "in a future release. Please use `update_item` instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        self.update_item(obj)


class ImageAxesParam(DataSet):
    """Parameters for the axes of an image plot."""

    xparams = BeginGroup(_("X Axis"))
    xmin = FloatItem("x|min", help=_("Lower x-axis limit"))
    xmax = FloatItem("x|max", help=_("Upper x-axis limit"))
    _xparams = EndGroup("end X")
    yparams = BeginGroup(_("Y Axis"))
    ymin = FloatItem("y|min", help=_("Lower y-axis limit"))
    ymax = FloatItem("y|max", help=_("Upper y-axis limit"))
    _yparams = EndGroup("end Y")
    zparams = BeginGroup(_("Z Axis"))
    zmin = FloatItem("z|min", help=_("Lower z-axis limit"))
    zmax = FloatItem("z|max", help=_("Upper z-axis limit"))
    _zparams = EndGroup("end Z")

    def update_param(self, item: BaseImageItem) -> None:
        """
        Update the parameters of the axes associated with the image item.

        Args:
            item: The image item from which to update the parameters.
        """
        plot: BasePlot = item.plot()
        xaxis: QwtScaleDiv = plot.axisScaleDiv(item.xAxis())
        self.xmin = xaxis.lowerBound()
        self.xmax = xaxis.upperBound()
        yaxis: QwtScaleDiv = plot.axisScaleDiv(item.yAxis())
        self.ymin = yaxis.lowerBound()
        self.ymax = yaxis.upperBound()
        self.zmin, self.zmax = item.min, item.max

    def update_item(self, item: BaseImageItem) -> None:
        """
        Update the axes with the parameters associated with the image item.

        Args:
            item: The image item to update.
        """
        plot: BasePlot = item.plot()
        plot.set_plot_limits(self.xmin, self.xmax, self.ymin, self.ymax)
        item.set_lut_range([self.zmin, self.zmax])
        plot.update_colormap_axis(item)
        for axis_id in (X_BOTTOM, Y_LEFT, Y_RIGHT):
            plot.SIG_AXIS_PARAMETERS_CHANGED.emit(axis_id)

    # TODO: remove this method in a future release
    def update_axes(self, obj: BaseImageItem) -> None:
        """Update object from parameters. Deprecated, use update_item instead.

        Args:
            obj: The image item to update.
        """
        warnings.warn(
            "`ImageAxesParam.update_axes` method is deprecated and will be removed "
            "in a future release. Please use `update_item` instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        self.update_item(obj)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

# pylint: disable-msg=W0622,R0903
# W0622: complains about filter being a builtin
# R0903: complains about too few public methods which is the purpose here

"""
Event handling
--------------

Overview
^^^^^^^^

The :mod:`plotpy.events` module provides classes to handle events on a
:class:`plotpy.plot.PlotWidget`.

The following classes are available:

* :class:`.StatefulEventFilter`: stateful event filter
* :class:`.KeyEventMatch`: key event match
* :class:`.StandardKeyMatch`: standard key event match
* :class:`.MouseEventMatch`: mouse event match
* :class:`.GestureEventMatch`: gesture event match
* :class:`.DragHandler`: drag handler
* :class:`.ClickHandler`: click handler
* :class:`.PanHandler`: pan handler
* :class:`.ZoomHandler`: zoom handler
* :class:`.GestureHandler`: gesture handler
* :class:`.PinchPanGestureHandler`: pinch and pan gesture handler
* :class:`.MenuHandler`: menu handler
* :class:`.QtDragHandler`: Qt drag handler
* :class:`.AutoZoomHandler`: auto zoom handler
* :class:`.MoveHandler`: move handler
* :class:`.ObjectHandler`: object handler
* :class:`.RectangularSelectionHandler`: rectangular selection handler
* :class:`.PointSelectionHandler`: point selection handler
* :class:`.ZoomRectHandler`: zoom rectangle handler

Reference
^^^^^^^^^

.. autoclass:: StatefulEventFilter
    :members:
.. autoclass:: KeyEventMatch
    :members:
.. autoclass:: StandardKeyMatch
    :members:
.. autoclass:: MouseEventMatch
    :members:
.. autoclass:: GestureEventMatch
    :members:
.. autoclass:: DragHandler
    :members:
.. autoclass:: ClickHandler
    :members:
.. autoclass:: PanHandler
    :members:
.. autoclass:: ZoomHandler
    :members:
.. autoclass:: GestureHandler
    :members:
.. autoclass:: PinchPanGestureHandler
    :members:
.. autoclass:: MenuHandler
    :members:
.. autoclass:: QtDragHandler
    :members:
.. autoclass:: AutoZoomHandler
    :members:
.. autoclass:: MoveHandler
    :members:
.. autoclass:: ObjectHandler
    :members:
.. autoclass:: RectangularSelectionHandler
    :members:
.. autoclass:: PointSelectionHandler
    :members:
.. autoclass:: ZoomRectHandler
    :members:
"""

from __future__ import annotations

import weakref
from typing import TYPE_CHECKING, Any, Callable

import numpy as np
from qtpy import QtCore as QC
from qtpy import QtGui as QG
from qtpy import QtWidgets as QW

from plotpy.config import CONF
from plotpy.constants import SHAPE_Z_OFFSET
from plotpy.coords import axes_to_canvas, canvas_to_axes
from plotpy.items import AnnotatedPolygon, Marker, PolygonShape

if TYPE_CHECKING:
    from qtpy.QtCore import QPoint
    from qtpy.QtGui import QMouseEvent

    from plotpy.plot.base import BasePlot


def buttons_to_str(buttons: int) -> str:
    """Conversion of Qt flags to string

    Args:
        buttons: Qt flags (e.g. `Qt.LeftButton`, `Qt.RightButton`, etc.)

    Returns:
        Human readable string for the button flags
    """
    string = ""
    if buttons & QC.Qt.LeftButton:
        string += "L"
    if buttons & QC.Qt.MidButton:  # Do not use QC.Qt.MouseButton.MidButton (Qt6!)
        string += "M"
    if buttons & QC.Qt.RightButton:
        string += "R"
    return string


def evt_type_to_str(type: int) -> str:
    """Convert event type to human readable string

    Args:
        type: event type

    Returns:
        Human readable string for the event type
    """
    if type == QC.QEvent.MouseButtonPress:
        return "Mpress"
    elif type == QC.QEvent.MouseButtonRelease:
        return "Mrelease"
    elif type == QC.QEvent.MouseMove:
        return "Mmove"
    elif type == QC.QEvent.ContextMenu:
        return "Context"
    return f"{type:d}"


# MARK: Event handlers -----------------------------------------------------------------
class EventMatch:
    """A callable returning true if it matches an event"""

    def __call__(self, event: QC.QEvent) -> bool:
        """Returns True if the event matches the event match

        Args:
            event: event to match
        """
        raise NotImplementedError

    def get_event_types(self) -> frozenset[int]:
        """Returns a set of event types handled by this
        EventMatch.

        This is used to quickly optimize events not handled by any event matchers
        """
        return frozenset()


class KeyEventMatch(EventMatch):
    """
    A callable returning True if it matches a key event

    Args:
        keys: list of keys or couples (key, modifier)
    """

    def __init__(self, keys: list[int | tuple[int, int]]) -> None:
        super().__init__()
        key_list, mod_list = [], []
        for item in keys:
            if isinstance(item, (tuple, list)):
                k, m = item
            else:
                k = item
                # Avoid bad arguments: modifier instead of key
                assert k not in (
                    QC.Qt.ControlModifier,
                    QC.Qt.ShiftModifier,
                    QC.Qt.AltModifier,
                    QC.Qt.NoModifier,
                )
                m = QC.Qt.NoModifier
            key_list.append(k)
            mod_list.append(m)
        self.keys = key_list
        self.mods = mod_list

    def get_event_types(self) -> frozenset[int]:
        """Return the set of event types handled by this event match"""
        return frozenset((QC.QEvent.KeyPress,))

    def __call__(self, event: QG.QKeyEvent) -> bool:
        """Returns True if the event matches the event match

        Args:
            event: event to match
        """
        if event.type() == QC.QEvent.KeyPress:
            my_key = event.key()
            my_mod = event.modifiers()
            if my_key in self.keys:
                mod = self.mods[self.keys.index(my_key)]
                if mod == QC.Qt.NoModifier or my_mod & mod:
                    return True
        return False


class StandardKeyMatch(EventMatch):
    """
    A callable returning True if it matches a key event

    Args:
        keysequence: QKeySequence.StandardKey integer
    """

    def __init__(self, keysequence: int) -> None:
        super().__init__()
        assert isinstance(keysequence, (int, QG.QKeySequence.StandardKey))
        self.keyseq = keysequence

    def get_event_types(self) -> frozenset[int]:
        """Return the set of event types handled by this event match"""
        return frozenset((QC.QEvent.KeyPress,))

    def __call__(self, event: QG.QKeyEvent) -> bool:
        """Returns True if the event matches the event match

        Args:
            event: event to match
        """
        return event.type() == QC.QEvent.KeyPress and event.matches(self.keyseq)


class MouseEventMatch(EventMatch):
    """Base class for matching mouse events

    Args:
        evt_type: event type
        btn: button to match
        modifiers: keyboard modifiers to match
    """

    def __init__(
        self, evt_type: int, btn: int, modifiers: int = QC.Qt.NoModifier
    ) -> None:
        super().__init__()
        assert isinstance(modifiers, (int, QC.Qt.KeyboardModifier))
        self.evt_type = evt_type
        self.button = btn
        self.modifiers = modifiers

    def get_event_types(self) -> frozenset[int]:
        """Return the set of event types handled by this event match"""
        return frozenset((self.evt_type,))

    def __call__(self, event: QMouseEvent) -> bool:
        """Returns True if the event matches the event match

        Args:
            event: event to match
        """
        if event.type() == self.evt_type:
            if event.button() == self.button:
                if self.modifiers != QC.Qt.NoModifier:
                    if (event.modifiers() & self.modifiers) == self.modifiers:
                        return True
                elif event.modifiers() == QC.Qt.NoModifier:
                    return True
        return False

    def __repr__(self) -> str:
        """Return textual representation"""
        return "<MouseMatch: {}/ {:08x}:{}>".format(
            evt_type_to_str(self.evt_type), self.modifiers, buttons_to_str(self.button)
        )


class MouseMoveMatch(MouseEventMatch):
    def __call__(self, event: QMouseEvent) -> bool:
        """Returns True if the event matches the event match

        Args:
            event: event to match
        """
        if event.type() == self.evt_type:
            if (
                self.button != QC.Qt.NoButton
                and (event.buttons() & self.button == self.button)
            ) or event.buttons() == QC.Qt.NoButton:
                if self.modifiers != QC.Qt.NoModifier:
                    if (event.modifiers() & self.modifiers) == self.modifiers:
                        return True
                elif event.modifiers() == QC.Qt.NoModifier:
                    return True
        return False


class GestureEventMatch(EventMatch):
    """Base class for matching gesture events"""

    def __init__(self, gesture_type: int, gesture_state: int) -> None:
        super().__init__()
        self.gesture_type = gesture_type
        self.gesture_state = gesture_state

    @staticmethod
    def __get_type_str(gesture_type: int) -> str:
        """Return text representation for gesture type"""
        for attr in (
            "TapGesture",
            "TapAndHoldGesture",
            "PanGesture",
            "PinchGesture",
            "SwipeGesture",
            "CustomGesture",
        ):
            if gesture_type == getattr(QC.Qt, attr):
                return attr

    @staticmethod
    def __get_state_str(gesture_state: int) -> str:
        """Return text representation for gesture state"""
        for attr in (
            "GestureStarted",
            "GestureUpdated",
            "GestureFinished",
            "GestureCanceled",
        ):
            if gesture_state == getattr(QC.Qt, attr):
                return attr

    def get_event_types(self) -> frozenset[int]:
        """Return the set of event types handled by this event match"""
        return frozenset((QC.QEvent.Gesture,))

    def __call__(self, event: QC.QEvent) -> bool:
        """Returns True if the event matches the event match

        Args:
            event: event to match
        """
        if event.type() == QC.QEvent.Gesture:
            gesture = event.gesture(self.gesture_type)
            return gesture and gesture.state() == self.gesture_state
        return False

    def __repr__(self) -> str:
        """Return textual representation"""
        type_str = self.__get_type_str(self.gesture_type)
        state_str = self.__get_state_str(self.gesture_state)
        return "<GestureMatch: %s:%s>" % (type_str, state_str)


class WheelEventMatch(EventMatch):
    """A callable returning True if it matches a wheel event

    Args:
        modifiers: keyboard modifiers
    """

    def __init__(self, modifiers: int = QC.Qt.KeyboardModifier.NoModifier) -> None:
        super().__init__()
        self.modifiers = modifiers

    def get_event_types(self) -> frozenset[int]:
        """Return the set of event types handled by this event match"""
        return frozenset((QC.QEvent.Type.Wheel,))

    def __call__(self, event: QG.QWheelEvent) -> bool:
        """Returns True if the event matches the event match

        Args:
            event: event to match
        """
        return isinstance(event, QG.QWheelEvent) and event.modifiers() == self.modifiers


# MARK: Finite state machine -----------------------------------------------------------
class StatefulEventFilter(QC.QObject):
    """State machine for handling events of a plot's canvas

    Args:
        parent: plot on which to install the event filter
    """

    def __init__(self, parent: BasePlot) -> None:
        super().__init__()

        # Machine states: (0: cursor, 1: panning, 2: zooming)
        self.states: dict[int, dict[EventMatch, list[list[callable], int]]] = {0: {}}

        self.cursors: dict[int, QC.Qt.CursorShape] = {}
        self.state = 0
        self.max_state = 0
        self.events: dict[tuple[str, int, int], EventMatch] = {}
        self.plot: BasePlot = parent
        self.all_event_types = frozenset()

    def eventFilter(self, _obj: QC.QObject, event: QC.QEvent) -> bool:
        """The `eventfilter` callback for Qt"""
        if not hasattr(self, "all_event_types"):
            print(repr(self), self)
        if event.type() not in self.all_event_types:
            return False
        state = self.states[self.state]
        for match, (call_list, next_state) in list(state.items()):
            if match(event):
                self.set_state(next_state, event)
                for call in call_list:
                    call(self, event)  # might change state
        return False

    def set_state(self, state: int, event: QC.QEvent) -> None:
        """Change the current state.
        Can be called by handlers to cancel a state change

        Args:
            state: new state
            event: event that triggered the state change
        """
        assert state in self.states
        if state == self.state:
            return
        self.state = state
        cursor = self.get_cursor(event)
        if cursor is not None:
            self.plot.canvas().setCursor(cursor)

    def new_state(self) -> int:
        """Create (reserve) a new state number

        Returns:
            New state number
        """
        self.max_state += 1
        self.states[self.max_state] = {}
        return self.max_state

    def add_event(
        self, state: int, match: EventMatch, call: callable, next_state: int = None
    ) -> int:
        """Add a transition to the state machine.
        If next_state is provided, it must correspond to an existing state,
        otherwise a new destination state is created

        Args:
            state: current state
            match: event matcher
            call: callable to execute when the event matches
            next_state: next state to go to when the event matches

        Returns:
            Next state number
        """
        assert isinstance(state, int)
        assert isinstance(match, EventMatch)
        self.all_event_types = self.all_event_types.union(match.get_event_types())
        entry = self.states[state].setdefault(match, [[], None])
        entry[0].append(call)
        if entry[1] is None:
            if next_state is None:
                next_state = self.new_state()
            else:
                pass
        else:
            if next_state is not None:
                assert next_state == entry[1]
            else:
                next_state = entry[1]

        entry[1] = next_state
        return next_state

    def nothing(self, filter: StatefulEventFilter, event: QC.QEvent) -> None:
        """A nothing filter, provided to help removing duplicate handlers

        Args:
            filter: event filter that contains the BasePlot instance
            event: event that triggered the action
        """
        pass

    # Cursor management ----------------------------------------------------------------
    def set_cursor(self, cursor: QC.Qt.CursorShape, *states: int) -> None:
        """Associate a cursor with one or more states

        Args:
            cursor: cursor shape
            states: state numbers
        """
        assert isinstance(cursor, QC.Qt.CursorShape)
        for s in states:
            self.cursors[s] = cursor

    def get_cursor(self, _event: QC.QEvent) -> QC.Qt.CursorShape | None:
        """Get the cursor associated with a given state/event

        Returns:
            Cursor shape or None if no cursor is associated with the state
        """
        # pass event to potentially choose the cursor based on modifier keys
        cursor = self.cursors.get(self.state, None)
        if cursor is None:
            # no cursor specified : should keep previous one
            return None
        return cursor

    # Mouse management -----------------------------------------------------------------
    def mouse_press(
        self, btn: int, modifiers: int = QC.Qt.NoModifier
    ) -> MouseEventMatch:
        """Create a filter for the MousePress event

        Args:
            btn: button to match
            modifiers: keyboard modifiers to match

        Returns:
            An instance of MouseEventMatch that matches the MousePress event
        """
        return self.events.setdefault(
            ("mousepress", btn, modifiers),
            MouseEventMatch(QC.QEvent.MouseButtonPress, btn, modifiers),
        )

    def mouse_move(
        self, btn: int, modifiers: int = QC.Qt.NoModifier
    ) -> MouseEventMatch:
        """Create a filter for the MouseMove event

        Args:
            btn: button to match
            modifiers: keyboard modifiers to match

        Returns:
            An instance of MouseEventMatch that matches the MouseMove event
        """
        return self.events.setdefault(
            ("mousemove", btn, modifiers),
            MouseMoveMatch(QC.QEvent.MouseMove, btn, modifiers),
        )

    def mouse_release(
        self, btn: int, modifiers: int = QC.Qt.NoModifier
    ) -> MouseEventMatch:
        """Create a filter for the MouseRelease event

        Args:
            btn: The button to match
            modifiers: The keyboard modifiers to match

        Returns:
            An instance of MouseEventMatch that matches the MouseRelease event
        """
        return self.events.setdefault(
            ("mouserelease", btn, modifiers),
            MouseEventMatch(QC.QEvent.MouseButtonRelease, btn, modifiers),
        )

    # Gesture management ---------------------------------------------------------------
    def gesture(self, kind: int, state: int) -> GestureEventMatch:
        """Create a filter for the gesture event

        Args:
            kind: The type of gesture
            state: The state of the gesture

        Returns:
            An instance of GestureEventMatch that matches the gesture event
        """
        return self.events.setdefault(
            ("gesture", kind, state), GestureEventMatch(kind, state)
        )

    # Wheel management -----------------------------------------------------------------
    def wheel(
        self, modifiers: int = QC.Qt.KeyboardModifier.NoModifier
    ) -> WheelEventMatch:
        """Create a filter for wheel events

        Args:
            modifiers: The keyboard modifiers to use with the wheel event

        Returns:
            An instance of WheelEventMatch that matches the wheel event"""
        return self.events.setdefault(("wheel", modifiers), WheelEventMatch(modifiers))


# MARK: Event handlers -----------------------------------------------------------------
class DragHandler(QC.QObject):
    """Base class for click-drag-release event handlers.

    Args:
        filter: The StatefulEventFilter instance.
        btn: The mouse button to match.
        mods: The keyboard modifiers to match. (default: QC.Qt.NoModifier)
        start_state: The starting state. (default: 0)
    """

    cursor = None

    def __init__(
        self,
        filter: StatefulEventFilter,
        btn: int,
        mods: int = QC.Qt.NoModifier,
        start_state: int = 0,
    ) -> None:
        super().__init__()
        self.state0 = filter.add_event(
            start_state, filter.mouse_press(btn, mods), self.start_tracking
        )
        self.state1 = filter.add_event(
            self.state0, filter.mouse_move(btn, mods), self.start_moving
        )
        filter.add_event(
            self.state1, filter.mouse_move(btn, mods), self.move, self.state1
        )
        filter.add_event(
            self.state0,
            filter.mouse_release(btn, mods),
            self.stop_notmoving,
            start_state,
        )
        filter.add_event(
            self.state1, filter.mouse_release(btn, mods), self.stop_moving, start_state
        )
        if self.cursor is not None:
            filter.set_cursor(self.cursor, self.state0, self.state1)
        self.start = None  # first mouse position
        self.last = None  # mouse position seen during last event
        self.parent_tracking = None

    def get_move_state(
        self, filter: StatefulEventFilter, pos: QC.QPointF
    ) -> tuple[float, float]:
        """Get the movement state based on the current filter and position

        Args:
            filter: The StatefulEventFilter instance
            pos: The current position

        Returns:
            A tuple containing the movement state in the x and y directions
        """
        rct = filter.plot.contentsRect()
        dx = (pos.x(), self.last.x(), self.start.x(), rct.width())
        dy = (pos.y(), self.last.y(), self.start.y(), rct.height())
        self.last = QC.QPointF(pos)
        return dx, dy

    def start_tracking(self, _filter: StatefulEventFilter, event: QMouseEvent) -> None:
        """Start tracking the mouse movement

        Args:
            _filter: The StatefulEventFilter instance
            event: The mouse event
        """
        self.start = self.last = QC.QPointF(event.pos())

    def start_moving(self, filter: StatefulEventFilter, event: QMouseEvent) -> None:
        """Start moving the mouse

        Args:
            filter: The StatefulEventFilter instance
            event: The mouse event
        """
        return self.move(filter, event)

    def stop_tracking(self, _filter: StatefulEventFilter, _event: QMouseEvent) -> None:
        """Stop tracking the mouse movement

        Args:
            _filter: The StatefulEventFilter instance
            _event: The mouse event
        """
        pass
        # filter.plot.canvas().setMouseTracking(self.parent_tracking)

    def stop_notmoving(self, filter: StatefulEventFilter, event: QMouseEvent) -> None:
        """Stop tracking the mouse movement when the mouse is not moving.

        Args:
            filter: The StatefulEventFilter instance.
            event: The mouse event.
        """
        self.stop_tracking(filter, event)

    def stop_moving(self, filter: StatefulEventFilter, event: QMouseEvent) -> None:
        """Stop moving the mouse

        Args:
            filter: The StatefulEventFilter instance
            event: The mouse event
        """
        self.stop_tracking(filter, event)

    def move(self, filter: StatefulEventFilter, event: QMouseEvent) -> None:
        """Move the mouse

        Args:
            filter: The StatefulEventFilter instance
            event: The mouse event
        """
        raise NotImplementedError()


class ClickHandler(QC.QObject):
    """Base class for click-release event handlers.

    Args:
        filter: The StatefulEventFilter instance.
        btn: The mouse button to match.
        mods: The keyboard modifiers to match. (default: QC.Qt.NoModifier)
        start_state: The starting state. (default: 0)

    Signals:
        SIG_CLICK_EVENT: Signal emitted by ClickHandler on mouse click
    """

    #: Signal emitted by ClickHandler on mouse click
    SIG_CLICK_EVENT = QC.Signal(object, "QEvent")

    def __init__(
        self,
        filter: StatefulEventFilter,
        btn: int,
        mods: int = QC.Qt.NoModifier,
        start_state: int = 0,
    ) -> None:
        super().__init__()
        self.state0 = filter.add_event(
            start_state, filter.mouse_press(btn, mods), filter.nothing
        )
        filter.add_event(
            self.state0, filter.mouse_release(btn, mods), self.click, start_state
        )

    def click(self, filter: StatefulEventFilter, event: QC.QEvent) -> None:
        """Handle the click event.

        Args:
            filter: The StatefulEventFilter instance.
            event: The mouse event.
        """
        self.SIG_CLICK_EVENT.emit(filter, event)


class PanHandler(DragHandler):
    cursor = QC.Qt.ClosedHandCursor

    def move(self, filter: StatefulEventFilter, event: QMouseEvent) -> None:
        """Move the mouse.

        Args:
            filter: The StatefulEventFilter instance.
            event: The mouse event.
        """
        x_state, y_state = self.get_move_state(filter, QC.QPointF(event.pos()))
        filter.plot.do_pan_view(x_state, y_state)


class ZoomHandler(DragHandler):
    cursor = QC.Qt.SizeAllCursor

    def move(self, filter: StatefulEventFilter, event: QMouseEvent) -> None:
        """Move the mouse.

        Args:
            filter: The StatefulEventFilter instance.
            event: The mouse event.
        """
        x_state, y_state = self.get_move_state(filter, QC.QPointF(event.pos()))
        filter.plot.do_zoom_view(x_state, y_state)


class GestureHandler(QC.QObject):
    """Base class used to handle gestures.

    Args:
        filter: event filter into which to add the handler instance.
        start_state: start state to use in the event filter state machine.
         Defaults to 0.
    """

    kind: QC.Qt.GestureType | None = None

    def __init__(self, filter: StatefulEventFilter, start_state: int = 0) -> None:
        super().__init__()
        filter.plot.canvas().grabGesture(self.kind)
        self.state0 = filter.add_event(
            start_state,
            filter.gesture(self.kind, QC.Qt.GestureState.GestureStarted),
            self.start_tracking,
        )
        self.state1 = filter.add_event(
            self.state0,
            filter.gesture(self.kind, QC.Qt.GestureState.GestureUpdated),
            self.start_moving,
        )
        filter.add_event(
            self.state1,
            filter.gesture(self.kind, QC.Qt.GestureState.GestureUpdated),
            self.move,
            self.state1,
        )
        filter.add_event(
            self.state0,
            filter.gesture(self.kind, QC.Qt.GestureState.GestureFinished),
            self.stop_tracking,
            start_state,
        )
        filter.add_event(
            self.state1,
            filter.gesture(self.kind, QC.Qt.GestureState.GestureFinished),
            self.stop_tracking,
            start_state,
        )
        filter.add_event(
            self.state0,
            filter.gesture(self.kind, QC.Qt.GestureState.GestureCanceled),
            self.stop_tracking,
            start_state,
        )
        filter.add_event(
            self.state1,
            filter.gesture(self.kind, QC.Qt.GestureState.GestureCanceled),
            self.stop_tracking,
            start_state,
        )
        self.start = QC.QPoint()  # first gesture position
        self.last = QC.QPoint()  # gesture position seen during last event
        self.parent_tracking = None

    def get_glob_position(self, event: QW.QGestureEvent) -> QC.QPointF:
        """Returns the hotspot global position of the gesture event.

        Args:
            event: event from which to get the position

        Returns:
            event hotspot poisition in global coordinates.
        """
        return self.get_gesture(event).hotSpot()

    def get_gesture(self, event: QW.QGestureEvent) -> QW.QGesture:
        """Returns the gesture from the event.

        Args:
            event: event from which to get the gesture.

        Returns:
            gesture from the event.
        """
        return event.gesture(self.kind)

    def start_tracking(
        self, filter: StatefulEventFilter, event: QW.QGestureEvent
    ) -> None:
        """Handles the start of the gesture tracking.

        Args:
            filter: event filter that contains the BasePlot instance
            event: event that triggered the start of the tracking, used to get
             a position.
        """
        origin = self.get_glob_position(event)
        self.start = self.last = filter.plot.canvas().mapFromGlobal(origin.toPoint())

    def start_moving(
        self, filter: StatefulEventFilter, event: QW.QGestureEvent
    ) -> None:
        """Handles the start of the gesture moving.

        Args:
            filter: event filter that contains the BasePlot instance
            event: event that triggered the start of the moving
        """
        return self.move(filter, event)

    def stop_tracking(
        self, _filter: StatefulEventFilter, _event: QW.QGestureEvent
    ) -> None:
        """Handles the stop of the gesture tracking.

        Args:
            _filter: event filter that contains the BasePlot instance
            _event: event that triggered the stop of the tracking
        """
        pass

    def move(self, filter: StatefulEventFilter, event: QW.QGestureEvent) -> None:
        """Handles the movement of the gesture.

        Args:
            filter: event filter that contains the BasePlot instance
            event: event that triggered the movement
        """
        raise NotImplementedError


class PinchPanGestureHandler(GestureHandler):
    """Class used to handle pinch and pan gestures.

    Args:
        filter: event filter into which to add the handler instance.
        start_state: start state to use in the event filter state machine.
         Defaults to 0.
    """

    kind = QC.Qt.GestureType.PinchGesture

    def __init__(
        self,
        filter: StatefulEventFilter,
        start_state=0,
    ) -> None:
        super().__init__(filter, start_state)
        self.last_center_diff = 0
        self.marker: Marker | None = None

    def get_pan_param(
        self, plot: BasePlot, pos: QPoint
    ) -> tuple[tuple[float, float, float, float], tuple[float, float, float, float]]:
        """Returns the parameters to use for panning the plot.

        Args:
            plot: instance of BasePlot to use as a reference.
            pos: position on the plot canvas of the current hotspot.

        Returns:
            Returns two tuples of four floats each, representing the parameters used
            by BasePlot.do_pan_view.
        """
        rct = plot.contentsRect()
        dx = (pos.x(), self.last.x(), self.start.x(), rct.width())
        dy = (pos.y(), self.last.y(), self.start.y(), rct.height())
        self.last = pos
        return dx, dy

    def get_zoom_param(
        self, plot: BasePlot, pos: QPoint, factor: float
    ) -> tuple[tuple[float, float, float, float], tuple[float, float, float, float]]:
        """Returns the parameters to use for zooming on the plot.

        Args:
            plot: instance of BasePlot to use as a reference.
            pos: position on the plot canvas of the current hotspot.
            factor: factor by which to zoom (zero-centered).

        Returns:
            Returns two tuples of four floats each, representing the parameters used
            by BasePlot.do_zoom_view.
        """
        rect = plot.contentsRect()
        w, h = rect.width(), rect.height()
        x, y = pos.x(), pos.y()
        dx = (
            x + (w * factor),
            x,
            x,
            w,
        )
        dy = (
            y + (h * factor),
            y,
            y,
            h,
        )
        return dx, dy

    def start_tracking(
        self, filter: StatefulEventFilter, event: QW.QGestureEvent
    ) -> None:
        """Overrides the GestureHandler.start_tracking method to add a marker at the
        hostpot of the pinch gesture.

        Args:
            filter: event filter that contains the BasePlot instance
            event: Gesture event that triggered the start of the tracking, used to get
             a position.
        """
        plot = filter.plot
        if self.marker is None:
            self.marker = Marker()
            self.marker.attach(plot)
            self.marker.setZ(plot.get_max_z() + 1)
            self.marker.setVisible(True)
        pos = plot.canvas().mapFromGlobal(self.get_glob_position(event).toPoint())
        self.marker.move_local_point_to(0, pos)

        return super().start_tracking(filter, event)

    def move(self, filter: StatefulEventFilter, event: QW.QGestureEvent) -> None:
        """Overrides the GestureHandler.move method to handle the pinch and pan gesture.

        Args:
            filter: event filter that contains the BasePlot instance
            event: event that triggered the move, used to get the hotspot position.
        """
        gesture: QW.QPinchGesture = self.get_gesture(event)
        plot = filter.plot

        center_point = self.get_glob_position(event)
        center_point = filter.plot.canvas().mapFromGlobal(center_point.toPoint())
        scale_factor = np.clip(gesture.scaleFactor(), 0.95, 1.05) - 1

        pan_dx, pan_dy = self.get_pan_param(plot, center_point)
        zoom_dx, zoom_dy = self.get_zoom_param(plot, center_point, scale_factor)
        plot.do_pan_view(pan_dx, pan_dy, replot=True)
        plot.do_zoom_view(zoom_dx, zoom_dy, lock_aspect_ratio=True)

    def stop_tracking(
        self, _filter: StatefulEventFilter, _event: QW.QGestureEvent
    ) -> None:
        """Overrides the GestureHandler.stop_tracking method to remove the marker when
        the gesture tracking stops.

        Args:
            _filter: event filter that contains the BasePlot instance
            _event: event that triggered the stop, not used.
        """
        if self.marker is not None:
            self.marker.detach()
            self.marker = None


class MenuHandler(ClickHandler):
    """Class to handle context menu events."""

    def click(self, filter: StatefulEventFilter, event: QC.QEvent) -> None:
        """
        Handles the click event.

        Args:
            filter: The StatefulEventFilter instance.
            event: The QC.QEvent instance.
        """
        menu = filter.plot.get_context_menu()
        if menu:
            menu.popup(event.globalPos())


class WheelHandler(QC.QObject):
    """Base class for handling mouse wheel events.

    Args:
        filter: event filter into which to add the handler instance.
        mods: Keyboard modifier that can be used at the same time as the wheel to
         trigger the event. Defaults to QC.Qt.KeyboardModifier.NoModifier.
        start_state: start state to use in the event filter state machine.
         Defaults to 0."""

    def __init__(
        self,
        filter: StatefulEventFilter,
        mods=QC.Qt.KeyboardModifier.NoModifier,
        start_state=0,
    ) -> None:
        super().__init__()
        self.state0 = filter.add_event(
            start_state, filter.wheel(mods), self.wheel, start_state
        )

    def wheel(self, filter: StatefulEventFilter, event: QG.QWheelEvent) -> None:
        """Handles the wheel event.

        Args:
            filter: event filter that contains the BasePlot instance
            event: the wheel event that triggered the action."""
        raise NotImplementedError()


class WheelZoomHandler(WheelHandler):
    """Class to handle zooming with the mouse wheel."""

    def get_zoom_param(
        self, plot: BasePlot, pos: QPoint, factor: float
    ) -> tuple[tuple[float, float, float, float], tuple[float, float, float, float]]:
        """Returns the parameters to use for zooming on the plot.

        Args:
            plot: instance of BasePlot to use as a reference.
            pos: position on the plot canvas of the current hotspot.
            factor: factor by which to zoom (zero-centered).

        Returns:
            Returns two tuples of four floats each, representing the parameters used
            by BasePlot.do_zoom_view.
        """

        x, y = pos.x(), pos.y()
        rect = plot.contentsRect()
        w, h = rect.width(), rect.height()
        dx = (
            x + (w * factor),
            x,
            x,
            w,
        )
        dy = (
            y + (h * factor),
            y,
            y,
            h,
        )
        return dx, dy

    def wheel(self, filter: StatefulEventFilter, event: QG.QWheelEvent) -> None:
        """Overrides the WheelHandler.wheel method to handle the zooming with the mouse
        wheel.

        Args:
            filter: event filter that contains the BasePlot instance
            event: the wheel event that triggered the zooming. Use to get zoom factor
             and position
        """
        plot = filter.plot
        center_point = event.globalPos()
        center_point = filter.plot.canvas().mapFromGlobal(center_point)  # type: ignore
        zoom_factor = (event.angleDelta().y() / 120) * 0.08
        dx, dy = self.get_zoom_param(plot, center_point, zoom_factor)
        plot.do_zoom_view(dx, dy, lock_aspect_ratio=True)


class QtDragHandler(DragHandler):
    """Class to handle drag events using Qt signals."""

    #: Signal emitted by QtDragHandler when starting tracking
    SIG_START_TRACKING = QC.Signal(object, "QMouseEvent")

    #: Signal emitted by QtDragHandler when stopping tracking and not moving
    SIG_STOP_NOT_MOVING = QC.Signal(object, "QMouseEvent")

    #: Signal emitted by QtDragHandler when stopping tracking and moving
    SIG_STOP_MOVING = QC.Signal(object, "QMouseEvent")

    #: Signal emitted by QtDragHandler when moving
    SIG_MOVE = QC.Signal(object, "QMouseEvent")

    def start_tracking(self, filter: StatefulEventFilter, event: QMouseEvent) -> None:
        """Starts tracking the drag event.

        Args:
            filter: The StatefulEventFilter instance.
            event: The QC.QEvent instance.
        """
        DragHandler.start_tracking(self, filter, event)
        self.SIG_START_TRACKING.emit(filter, event)

    def stop_notmoving(self, filter: StatefulEventFilter, event: QMouseEvent) -> None:
        """Stops tracking when the drag event is not moving.

        Args:
            filter: The StatefulEventFilter instance.
            event: The QC.QEvent instance.
        """
        self.SIG_STOP_NOT_MOVING.emit(filter, event)

    def stop_moving(self, filter: StatefulEventFilter, event: QMouseEvent) -> None:
        """Stops the movement of the drag event.

        Args:
            filter: The StatefulEventFilter instance.
            event: The QC.QEvent instance
        """
        self.SIG_STOP_MOVING.emit(filter, event)

    def move(self, filter: StatefulEventFilter, event: QMouseEvent) -> None:
        """Handles the move event.

        Args:
            filter: The StatefulEventFilter instance.
            event: The QC.QEvent instance.
        """
        self.SIG_MOVE.emit(filter, event)


class AutoZoomHandler(ClickHandler):
    """Class to handle auto-zoom events."""

    def click(self, filter: StatefulEventFilter, _event: QC.QEvent) -> None:
        """Handles the click event.

        Args:
            filter: The StatefulEventFilter instance.
            _event: The QC.QEvent instance.
        """
        filter.plot.do_autoscale()


class MoveHandler:
    """Class to handle moving events.

    Args:
        filter: The StatefulEventFilter instance.
        btn: The mouse button to match.
        mods: The keyboard modifiers to match. (default: QC.Qt.NoModifier)
        start_state: The starting state. (default: 0)
    """

    def __init__(
        self,
        filter: StatefulEventFilter,
        btn: QC.Qt.MouseButton = QC.Qt.NoButton,
        mods: QC.Qt.KeyboardModifiers = QC.Qt.NoModifier,
        start_state: int = 0,
    ):
        filter.add_event(
            start_state, filter.mouse_move(btn, mods), self.move, start_state
        )

    def move(self, filter: StatefulEventFilter, event: QW.QGestureEvent) -> None:
        """
        Handles the move event.

        Args:
            filter: The StatefulEventFilter instance.
            event: The QGestureEvent instance.
        """
        filter.plot.do_move_marker(event)


class UndoMoveObject:
    """Class to handle undo/redo events for moving objects.

    Args:
        obj: The object to move.
        pos1: The initial position.
        pos2: The final position.
    """

    def __init__(
        self, obj: Any, pos1: tuple[float, float], pos2: tuple[float, float]
    ) -> None:
        self.obj = obj
        self.coords1 = canvas_to_axes(obj, pos1)
        self.coords2 = canvas_to_axes(obj, pos2)

    def is_valid(self) -> bool:
        """Check if the object is valid.

        Returns:
            True if the object is valid, False otherwise.
        """
        return self.obj.plot() is not None

    def compute_positions(self) -> tuple[QC.QPointF, QC.QPointF]:
        """Compute the positions of the object.

        Returns:
            A tuple containing the positions of the object.
        """
        pos1 = QC.QPointF(*axes_to_canvas(self.obj, *self.coords1))
        pos2 = QC.QPointF(*axes_to_canvas(self.obj, *self.coords2))
        return pos1, pos2

    def undo(self) -> None:
        """Undo the action."""
        pos1, pos2 = self.compute_positions()
        self.obj.plot().unselect_all()
        self.obj.move_local_shape(pos1, pos2)

    def redo(self) -> None:
        """Redo the action."""
        pos1, pos2 = self.compute_positions()
        self.obj.plot().unselect_all()
        self.obj.move_local_shape(pos2, pos1)


class UndoMovePoint(UndoMoveObject):
    """Class to handle undo/redo events for moving points.

    Args:
        obj: The object to move.
        pos1: The initial position.
        pos2: The final position.
        handle: The handle to move.
        ctrl: The control point.
    """

    def __init__(
        self,
        obj: Any,
        pos1: tuple[float, float],
        pos2: tuple[float, float],
        handle: Any,
        ctrl: Any,
    ) -> None:
        super().__init__(obj, pos1, pos2)
        self.handle = handle
        self.ctrl = ctrl

    def undo(self) -> None:
        """Undo the action."""
        pos1, pos2 = self.compute_positions()
        self.obj.move_local_point_to(self.handle, pos1, self.ctrl)

    def redo(self) -> None:
        """Redo the action."""
        pos1, pos2 = self.compute_positions()
        self.obj.move_local_point_to(self.handle, pos2, self.ctrl)


class UndoRotatePoint(UndoMoveObject):
    """Class to handle undo/redo events for rotating points.

    Args:
        obj: The object to move.
        pos1: The initial position.
        pos2: The final position.
    """

    def __init__(
        self, obj: Any, pos1: tuple[float, float], pos2: tuple[float, float]
    ) -> None:
        super().__init__(obj, pos1, pos2)

    def undo(self) -> None:
        """Undo the action."""
        pos1, pos2 = self.compute_positions()
        self.obj.plot().unselect_all()
        self.obj.rotate_local_shape(pos1, pos2)

    def redo(self) -> None:
        """Redo the action."""
        pos1, pos2 = self.compute_positions()
        self.obj.plot().unselect_all()
        self.obj.rotate_local_shape(pos2, pos1)


class ObjectHandler:
    """Base class for handling objects.

    Args:
        filter: The StatefulEventFilter instance.
        btn: The mouse button to match.
        mods: The keyboard modifiers to match. (default: QC.Qt.NoModifier)
        start_state: The starting state. (default: 0)
        multiselection: Whether to allow multiple selections. (default: False)
    """

    def __init__(
        self,
        filter: StatefulEventFilter,
        btn: int,
        mods: QC.Qt.KeyboardModifiers = QC.Qt.NoModifier,
        start_state: int = 0,
        multiselection: bool = False,
    ) -> None:
        self.multiselection = multiselection
        self.start_state = start_state
        self.state0 = filter.add_event(
            start_state, filter.mouse_press(btn, mods), self.start_tracking
        )
        filter.add_event(
            self.state0, filter.mouse_move(btn, mods), self.move, self.state0
        )
        filter.add_event(
            self.state0,
            filter.mouse_release(btn, mods),
            self.stop_tracking,
            start_state,
        )
        filter.add_event(
            start_state, StandardKeyMatch(QG.QKeySequence.Undo), self.undo, start_state
        )
        filter.add_event(
            start_state, StandardKeyMatch(QG.QKeySequence.Redo), self.redo, start_state
        )
        self.handle = None  # first mouse position
        self.inside = False
        self.rotate_inside = False  # cas de la rotation avec un point à l'intérieur
        self._active = None  # mouse position seen during last event
        self.last_pos = None
        self.unselection_pending = None

        self.undo_stack = [None]
        self.undo_index = 0
        self.first_pos = None
        self.undo_action = None

    @property
    def active(self) -> Any:
        """Return active object"""
        if self._active is not None:
            return self._active()

    @active.setter
    def active(self, value: Any) -> None:
        """Set active object"""
        if value is None:
            self._active = None
        else:
            self._active = weakref.ref(value)

    def add_undo_move_action(self, undo_action: Any) -> None:
        """Add undo action to undo stack"""
        self.undo_stack = self.undo_stack[: self.undo_index + 1]
        self.undo_stack.append(undo_action)
        self.undo_index = len(self.undo_stack) - 1

    def undo(self, filter: StatefulEventFilter, event: QC.QEvent) -> None:
        """Undo the last action.

        Args:
            filter: The StatefulEventFilter instance.
            event: The event triggering the undo action.
        """
        action = self.undo_stack[self.undo_index]
        if action is not None:
            if action.is_valid():
                action.undo()
                filter.plot.replot()
            else:
                self.undo_stack.remove(action)
            self.undo_index -= 1

    def redo(self, filter: StatefulEventFilter, event: QC.QEvent) -> None:
        """Redo the last action.

        Args:
            filter: The StatefulEventFilter instance.
            event: The event triggering the redo action.
        """
        if self.undo_index < len(self.undo_stack) - 1:
            action = self.undo_stack[self.undo_index + 1]
            if action.is_valid():
                action.redo()
                filter.plot.replot()
                self.undo_index += 1
            else:
                self.undo_stack.remove(action)

    def start_tracking(self, filter: StatefulEventFilter, event: QC.QEvent) -> None:
        """Starts tracking the mouse movement for selection and interaction.

        Args:
            filter: The filter object.
            event: The mouse event.
        """
        plot = filter.plot
        self.inside = False
        self.rotate_inside = False
        self.active = None
        self.handle = None
        self.first_pos = pos = event.pos()
        self.last_pos = QC.QPointF(pos)
        selected = plot.get_active_item()
        distance = CONF.get("plot", "selection/distance", 6)

        (
            nearest,
            nearest_dist,
            nearest_handle,
            nearest_inside,
        ) = plot.get_nearest_object(pos, distance)
        if nearest is not None:
            # Is the nearest object the real deal?
            if not nearest.can_select() or nearest_dist >= distance:
                # Looking for the nearest object in z containing cursor position
                (
                    nearest,
                    nearest_dist,
                    nearest_handle,
                    nearest_inside,
                ) = plot.get_nearest_object_in_z(pos)

        # This will unselect active item only if it's not moved afterwards:
        self.unselection_pending = selected is nearest
        if selected and not self.multiselection:
            # An item is selected
            self.active = selected
            (dist, self.handle, self.inside, other_object) = self.active.hit_test(pos)
            if other_object is not None:
                # e.g. LegendBoxItem: 'other_object' is the selected curve
                plot.set_active_item(other_object)
                return
            if dist >= distance and not self.inside:
                # The following allows to move together selected items by
                # clicking inside any of them (instead of active item only)
                other_selitems = [
                    _it
                    for _it in plot.get_selected_items()
                    if _it is not self.active and _it.can_move()
                ]
                for selitem in other_selitems:
                    dist, handle, inside, _other = selitem.hit_test(pos)
                    if dist < distance or inside:
                        self.inside = inside
                        break
                else:
                    self.__unselect_objects(filter)
                    filter.set_state(self.start_state, event)
                    return
        else:
            # No item is selected
            self.active = nearest
            self.handle = nearest_handle
            self.inside = nearest_inside
            dist = nearest_dist
            if nearest is not None:
                plot.set_active_item(nearest)
                if not nearest.selected:
                    if not self.multiselection:
                        plot.unselect_all()
                    plot.select_item(nearest)

        # Eventually move or resize selected object:
        self.__move_or_resize_object(dist, distance, event, filter)
        plot.replot()

    def __unselect_objects(self, filter: StatefulEventFilter) -> None:
        """Unselect selected object*s*

        Args:
            filter: The StatefulEventFilter instance.
        """
        plot = filter.plot
        plot.unselect_all()
        plot.replot()

    def __move_or_resize_object(
        self,
        dist: float,
        distance: float,
        event: QC.QEvent,
        filter: StatefulEventFilter,
    ) -> None:
        """Move or resize the object

        Args:
            dist: The distance.
            distance: The distance.
            event: The mouse event.
            filter: The StatefulEventFilter instance.
        """
        if (
            self.active is not None
            and self.active.can_move() is False
            and self.inside
            and self.active.can_rotate()
        ):
            # Rotate inside
            self.inside = False
            self.rotate_inside = True
            return
        if (
            dist < distance
            and self.handle is not None
            and (self.active.can_resize() or self.active.can_rotate())
        ):
            # Resize / move handle
            self.inside = False
            return
        if self.inside and self.active.can_move():
            # Move object
            return
        # can't move, can't resize
        filter.set_state(self.start_state, event)
        if self.unselection_pending:
            self.__unselect_objects(filter)

    def move(self, filter: StatefulEventFilter, event: QC.QEvent) -> None:
        """Move the object.

        Args:
            filter: The StatefulEventFilter instance.
            event: The mouse event.
        """
        if self.active is None:
            return
        self.unselection_pending = False
        if self.rotate_inside:
            self.active.rotate_local_shape(self.last_pos, event.pos())
            self.undo_action = UndoRotatePoint(self.active, event.pos(), self.first_pos)
        elif self.inside:
            self.active.move_local_shape(self.last_pos, event.pos())
            self.undo_action = UndoMoveObject(self.active, event.pos(), self.first_pos)
        else:
            ctrl = event.modifiers() & QC.Qt.ControlModifier == QC.Qt.ControlModifier
            self.active.move_local_point_to(self.handle, event.pos(), ctrl)
            self.undo_action = UndoMovePoint(
                self.active, self.first_pos, event.pos(), self.handle, ctrl
            )
        self.last_pos = QC.QPointF(event.pos())
        filter.plot.replot()

    def stop_tracking(self, filter: StatefulEventFilter, event: QC.QEvent) -> None:
        """Stop tracking the object.

        Args:
            filter: The StatefulEventFilter instance.
            event: The mouse event.
        """
        self.add_undo_move_action(self.undo_action)
        if self.unselection_pending:
            self.__unselect_objects(filter)
        self.handle = None
        self.active = None
        self.inside = False
        self.rotate_inside = False


class RectangularSelectionHandler(DragHandler):
    """Base class for handling rectangular selections.

    Args:
        filter: The StatefulEventFilter instance.
        btn: The mouse button to match.
        mods: The keyboard modifiers to match. (default: QC.Qt.NoModifier)
        start_state: The starting state. (default: 0)
    """

    #: Signal emitted by RectangularSelectionHandler when ending selection
    SIG_END_RECT = QC.Signal(object, "QPointF", "QPointF")

    def __init__(
        self,
        filter: StatefulEventFilter,
        btn: int,
        mods: QC.Qt.KeyboardModifiers = QC.Qt.NoModifier,
        start_state: int = 0,
    ) -> None:
        super().__init__(filter, btn, mods, start_state)
        self.avoid_null_shape = False
        self.setup_shape_cb = None
        self.shape = None
        self.shape_h0 = None
        self.shape_h1 = None

    def set_shape(
        self,
        shape: Any,
        h0: Any,
        h1: Any,
        setup_shape_cb: Callable | None = None,
        avoid_null_shape: bool = False,
    ) -> None:
        """Set the shape.

        Args:
            shape: The shape.
            h0: The first handle.
            h1: The second handle.
            setup_shape_cb: The setup shape callback.
            avoid_null_shape: Whether to avoid null shape.
        """
        self.shape = shape
        self.shape_h0 = h0
        self.shape_h1 = h1
        self.setup_shape_cb = setup_shape_cb
        self.avoid_null_shape = avoid_null_shape

    def start_tracking(self, filter: StatefulEventFilter, event: QMouseEvent) -> None:
        """Start tracking the mouse movement for selection and interaction.

        Args:
            filter: The StatefulEventFilter instance.
            event: The mouse event.
        """
        self.start = self.last = QC.QPointF(event.pos())

    def start_moving(self, filter: StatefulEventFilter, event: QMouseEvent) -> None:
        """Start moving the object.

        Args:
            filter: The StatefulEventFilter instance.
            event: The mouse event.
        """
        self.shape.attach(filter.plot)
        self.shape.setZ(filter.plot.get_max_z() + 1)
        if self.avoid_null_shape:
            self.start -= QC.QPointF(1, 1)
        self.shape.move_local_point_to(self.shape_h0, self.start)
        self.shape.move_local_point_to(self.shape_h1, QC.QPointF(event.pos()))
        self.start_moving_action(filter, event)
        if self.setup_shape_cb is not None:
            self.setup_shape_cb(self.shape)
        self.shape.show()
        filter.plot.replot()

    def start_moving_action(
        self, filter: StatefulEventFilter, event: QMouseEvent
    ) -> None:
        """Start moving the object.

        Args:
            filter: The StatefulEventFilter instance.
            event: The mouse event.
        """
        pass

    def move(self, filter: StatefulEventFilter, event: QMouseEvent) -> None:
        """Move the object.

        Args:
            filter: The StatefulEventFilter instance.
            event: The mouse event.
        """
        self.shape.move_local_point_to(self.shape_h1, QC.QPointF(event.pos()))
        self.move_action(filter, event)
        filter.plot.replot()

    def move_action(self, filter: StatefulEventFilter, event: QMouseEvent) -> None:
        """Move the object.

        Args:
            filter: The StatefulEventFilter instance.
            event: The mouse event.
        """
        pass

    def stop_moving(self, filter: StatefulEventFilter, event: QMouseEvent) -> None:
        """Stop moving the object.

        Args:
            filter: The StatefulEventFilter instance.
            event: The mouse event.
        """
        self.shape.detach()
        self.stop_moving_action(filter, event)
        filter.plot.replot()

    def stop_moving_action(
        self, filter: StatefulEventFilter, event: QMouseEvent
    ) -> None:
        """Stop moving the object.

        Args:
            filter: The StatefulEventFilter instance.
            event: The mouse event.
        """
        self.SIG_END_RECT.emit(filter, self.start, QC.QPointF(event.pos()))


class PointSelectionHandler(RectangularSelectionHandler):
    """Class to handle point selections."""

    def stop_notmoving(self, filter: StatefulEventFilter, event: QMouseEvent) -> None:
        """Stop the point selection when the point is not moving.

        Args:
            filter: The StatefulEventFilter instance.
            event: The mouse event.
        """
        self.stop_moving(filter, event)


class ZoomRectHandler(RectangularSelectionHandler):
    """Class to handle zoom rectangles."""

    def stop_moving_action(
        self, filter: StatefulEventFilter, event: QMouseEvent
    ) -> None:
        """Stop moving the object.

        Args:
            filter: The StatefulEventFilter instance.
            event: The mouse event.
        """
        filter.plot.do_zoom_rect_view(self.start, QC.QPointF(event.pos()))


class MultilineSelectionHandler(DragHandler):
    """
    A handler for handling multi-line selections.

    This handler extends the DragHandler to handle multi-line selections.

    Args:
        filter: The StatefulEventFilter instance.
        btn: The mouse button to match.
        mods: The keyboard modifiers to match. (default: QC.Qt.NoModifier)
        start_state: The starting state. (default: 0)
        closed: Whether the polygon should be closed. (default: False)
    """

    SIG_END_POLYLINE = QC.Signal(object, np.ndarray)

    def __init__(
        self,
        filter: StatefulEventFilter,
        btn: int,
        mods: QC.Qt.KeyboardModifiers = QC.Qt.NoModifier,
        start_state: int = 0,
        closed: bool = False,
    ) -> None:
        super().__init__(filter, btn, mods, start_state)
        self.closed = closed
        filter.add_event(
            start_state,
            KeyEventMatch((QC.Qt.Key_Enter, QC.Qt.Key_Return, QC.Qt.Key_Space)),
            self.accept_polygonshape,
            start_state,
        )
        filter.add_event(
            start_state,
            KeyEventMatch((QC.Qt.Key_Backspace, QC.Qt.Key_Escape)),
            self.cancel_point,
            start_state,
        )
        self.init_pos: QC.QPointF | None = None
        self.shape: PolygonShape | AnnotatedPolygon | None = None
        self.current_handle: int | None = None
        self.setup_shape_cb: Callable | None = None

    def set_shape(
        self,
        shape: PolygonShape | AnnotatedPolygon,
        setup_shape_cb: Callable | None = None,
    ) -> None:
        """Set the shape.

        Args:
            shape: The shape.
            setup_shape_cb: The setup shape callback.
        """
        self.shape = shape
        self.setup_shape_cb = setup_shape_cb

    def start_tracking(self, filter: StatefulEventFilter, event: QMouseEvent) -> None:
        """Starts tracking the drag event.

        Args:
            filter: The StatefulEventFilter instance.
            event: The QC.QEvent instance.
        """
        if self.init_pos is None:
            self.init_pos = QC.QPointF(event.pos())

    def start_moving(self, filter: StatefulEventFilter, event: QMouseEvent) -> None:
        """Start moving the object.

        Args:
            filter: The StatefulEventFilter instance.
            event: The mouse event.
        """
        if self.init_pos is None:
            return
        if self.shape.plot() is None:
            self.shape.set_points(None)
            filter.plot.add_item_with_z_offset(self.shape, SHAPE_Z_OFFSET)
            self.shape.setZ(filter.plot.get_max_z() + 1)
            if self.setup_shape_cb is not None:
                self.setup_shape_cb(self.shape)
            self.shape.add_local_point(self.init_pos)
        self.current_handle = self.shape.add_local_point(QC.QPointF(event.pos()))
        self.shape.set_closed(self.closed and len(self.shape.get_points()) > 2)
        self.shape.show()
        filter.plot.replot()

    def move(self, filter: StatefulEventFilter, event: QMouseEvent) -> None:
        """
        Handle mouse move event to update the position of the last point.

        Args:
            filter: The plot filter.
            event: The mouse event.
        """
        self.shape.move_local_point_to(self.current_handle, QC.QPointF(event.pos()))
        filter.plot.replot()

    def stop_notmoving(self, filter: StatefulEventFilter, event: QMouseEvent) -> None:
        """Stops tracking when the drag event is not moving.

        Args:
            filter: The StatefulEventFilter instance.
            event: The QC.QEvent instance.
        """
        points = self.shape.get_points()
        if len(points) > 1:
            self.accept_polygonshape(filter, event)

    def stop_moving(self, filter: StatefulEventFilter, event: QMouseEvent) -> None:
        """Stops the movement of the drag event.

        Args:
            filter: The StatefulEventFilter instance.
            event: The QC.QEvent instance
        """
        pos = QC.QPointF(event.pos())
        if self.init_pos == pos:
            self.shape.del_point(-1)
        else:
            self.shape.move_local_point_to(self.current_handle, pos)
        filter.plot.replot()

    def accept_polygonshape(
        self, filter: StatefulEventFilter, event: QMouseEvent
    ) -> None:
        """Accept the polygon shape.

        Args:
            filter: The StatefulEventFilter instance.
            event: The mouse event.
        """
        if self.shape.plot() is None:
            return
        filter.plot.del_item(self.shape)
        self.init_pos = None
        self.current_handle = None
        self.SIG_END_POLYLINE.emit(filter, self.shape.get_points())

    def cancel_point(self, filter: StatefulEventFilter, event: QMouseEvent) -> None:
        """
        Cancel the last point or remove the shape if it has less than 3 points.

        Args:
            filter: The plot filter.
            event: The triggering event.
        """
        points = self.shape.get_points()
        if points is None or self.shape.plot() is None:
            return
        if len(points) <= 2:
            filter.plot.del_item(self.shape)
            self.init_pos = None
            self.shape.closed = False
        else:
            if self.current_handle:
                newh = self.shape.del_point(self.current_handle)
            else:
                newh = self.shape.del_point(-1)
            self.current_handle = newh
            self.shape.set_closed(self.closed and len(self.shape.get_points()) > 2)
        filter.plot.replot()


def setup_standard_tool_filter(filter: StatefulEventFilter, start_state: int) -> int:
    """Creation of standard filters (pan/zoom) on middle/right buttons

    Args:
        filter: The StatefulEventFilter instance.
        start_state: The starting state.

    Returns:
        The starting state.
    """
    # Middle button
    # Do not use QC.Qt.MouseButton.MidButton (Qt6!)
    PanHandler(filter, QC.Qt.MidButton, start_state=start_state)
    AutoZoomHandler(filter, QC.Qt.MidButton, start_state=start_state)

    # Right button
    ZoomHandler(filter, QC.Qt.MouseButton.RightButton, start_state=start_state)
    MenuHandler(filter, QC.Qt.MouseButton.RightButton, start_state=start_state)

    # Gestures
    PinchPanGestureHandler(filter, start_state=start_state)

    # Other events
    MoveHandler(filter, start_state=start_state)
    MoveHandler(
        filter, start_state=start_state, mods=QC.Qt.KeyboardModifier.ShiftModifier
    )
    MoveHandler(
        filter, start_state=start_state, mods=QC.Qt.KeyboardModifier.AltModifier
    )
    WheelZoomHandler(
        filter, start_state=start_state, mods=QC.Qt.KeyboardModifier.ControlModifier
    )
    return start_state
